import logging
from typing import AsyncGenerator, Iterator, Union
import os
import io

from pydub import AudioSegment

from achatbot.common.interface import ITts
from achatbot.common.session import Session
from achatbot.common.types import GTTS_SYNTHESIS_FILE, GTTSArgs, RECORDS_DIR
from .base import BaseTTS


class GTTS(BaseTTS, ITts):
    TAG = "tts_g"

    @classmethod
    def get_args(cls, **kwargs) -> dict:
        return {**GTTSArgs().__dict__, **kwargs}

    def __init__(self, **args) -> None:
        self.args = GTTSArgs(**args)
        self.file_path = os.path.join(RECORDS_DIR, GTTS_SYNTHESIS_FILE)

    async def _inference(
        self, session: Session, text: str, **kwargs
    ) -> AsyncGenerator[bytes, None]:
        from gtts import gTTS

        with io.BytesIO() as f:
            tts = gTTS(text=text, lang=self.args.language, tld=self.args.tld, slow=self.args.slow)
            tts.write_to_fp(f)
            f.seek(0)

            audio: AudioSegment = AudioSegment.from_mp3(f)
            if self.args.speed_increase != 1.0:
                audio = audio.speedup(
                    playback_speed=self.args.speed_increase,
                    chunk_size=self.args.chunk_size,
                    crossfade=self.args.crossfade_lenght,
                )
            audio_resampled = (
                audio.set_frame_rate(22050).set_channels(1).set_sample_width(2)
            )  # 16bit sample_width 16/8=2
            audio_data = audio_resampled.raw_data
            yield audio_data

    def get_voices(self):
        from gtts.lang import tts_langs

        voices = []
        languages = tts_langs()
        tlds = ["com", "com.au", "co.uk", "us", "ca", "co.in", "ie", "co.za"]

        for lang in languages.keys():
            for tld in tlds:
                voices.append(f"{lang}-{tld}")

        return voices



from dataclasses import dataclass
from typing import Optional, Union

from achatbot.types.speech.asr.base import ASRArgs
from faster_whisper.vad import VadOptions


@dataclass
class WhisperFasterASRArgs(ASRArgs):
    vad_filter: bool = False
    vad_parameters: Optional[Union[dict, VadOptions]] = None



import os
import logging

from time import perf_counter
import unittest
import uuid
from dotenv import load_dotenv

from src.common.interface import ILlmGenerator
from src.core.llm.base import BaseLLM
from src.common.logger import Logger
from src.common.session import Session
from src.common.types import MODELS_DIR, SessionCtx
from src.core.llm import LLMEnvInit

load_dotenv(override=True)

r"""
# flashinfer needs to define TORCH_CUDA_ARCH_LIST
export TORCH_CUDA_ARCH_LIST="7.5 8.0 8.6 8.7 8.9 9.0"

LLM_MODEL_NAME_OR_PATH=./models/Qwen/Qwen2.5-0.5B-Instruct \
    LLM_TAG=llm_transformers_generator \
    python -m unittest test.core.llm.test_generator.TestGenerator.test_generate

LLM_MODEL_NAME_OR_PATH=./models/Qwen/Qwen2.5-0.5B-Instruct \
    LLM_MODEL_PATH=./models/qwen2.5-0.5b-instruct-q8_0.gguf \
    LLM_TAG=llm_llamacpp_generator \
    python -m unittest test.core.llm.test_generator.TestGenerator.test_generate

LLM_MODEL_NAME_OR_PATH=./models/Qwen/Qwen2.5-0.5B-Instruct \
    LLM_TAG=llm_vllm_generator \
    python -m unittest test.core.llm.test_generator.TestGenerator.test_generate

LLM_MODEL_NAME_OR_PATH=./models/Qwen/Qwen2.5-0.5B-Instruct \
    LLM_TAG=llm_sglang_generator \
    python -m unittest test.core.llm.test_generator.TestGenerator.test_generate

LLM_MODEL_NAME_OR_PATH=./models/Qwen/Qwen2.5-0.5B-Instruct \
    LLM_TAG=llm_trtllm_generator \
    python -m unittest test.core.llm.test_generator.TestGenerator.test_generate

LLM_MODEL_NAME_OR_PATH=./models/Qwen/Qwen2.5-0.5B-Instruct-trtllm \
    LLM_TOKENIZER_PATH=/models/Qwen/Qwen2.5-0.5B-Instruct \
    LLM_DEBUG_MODE=1 \
    LLM_TAG=llm_trtllm_runner_generator \
    python -m unittest test.core.llm.test_generator.TestGenerator.test_generate
"""


class TestGenerator(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        cls.llm_tag = os.getenv("LLM_TAG", "llm_transformers_generator")
        cls.prompt = os.getenv("LLM_PROMPT", "what's your name?")
        cls.model_path = os.getenv("LLM_MODEL_NAME_OR_PATH", "./models/Qwen/Qwen2.5-0.5B")
        cls.tokenizer_path = os.getenv("LLM_TOKENIZER_PATH", cls.model_path)
        Logger.init(os.getenv("LOG_LEVEL", "debug").upper(), is_file=False)

    @classmethod
    def tearDownClass(cls):
        pass

    async def asyncSetUp(self):
        pass

    def setUp(self):
        from transformers import AutoTokenizer, GenerationConfig

        engine = LLMEnvInit.initLLMEngine(self.llm_tag)
        self.assertIsInstance(engine, BaseLLM)
        self.assertIsInstance(engine, ILlmGenerator)
        self.engine: ILlmGenerator = engine
        logging.debug(self.engine.args)

        self.tokenizer = AutoTokenizer.from_pretrained(self.tokenizer_path)
        self.generation_config = {}
        if os.path.exists(os.path.join(self.model_path, "generation_config.json")):
            self.generation_config = GenerationConfig.from_pretrained(
                self.model_path, "generation_config.json"
            ).to_dict()

        self.session = Session(**SessionCtx(f"{self.llm_tag}_" + str(uuid.uuid4().hex)).__dict__)

    def tearDown(self):
        pass

    async def asyncTearDown(self):
        pass

    async def test_generate(self):
        prompt_cases = [
            {"prompt": "hello, my name is", "kwargs": {"max_new_tokens": 30, "stop_ids": []}},
            {
                "prompt": "hello, my name is",
                "kwargs": {"max_new_tokens": 30, "stop_ids": [13]},
            },  # prefill cache token test (trtllm default no cache, vllm and sglang default cache)
            {
                "prompt": "hello, what your name?",
                "kwargs": {"max_new_tokens": 30, "stop_ids": [13]},
            },
        ]

        for case in prompt_cases:
            with self.subTest(case=case):
                tokens = self.tokenizer(case["prompt"])
                self.session.ctx.state["token_ids"] = tokens["input_ids"]
                # gen_kwargs = {**self.generation_config, **case["kwargs"], **tokens}# hack test,
                # trtllm_runner, vllm  generate have some bug :)
                gen_kwargs = {**case["kwargs"], **tokens}
                logging.debug(gen_kwargs)
                iter = self.engine.generate(self.session, **gen_kwargs)

                generated_token_ids = []
                times = []
                start_time = perf_counter()
                async for item in iter:
                    # print(item)
                    generated_token_ids.append(item)
                    times.append(perf_counter() - start_time)
                    start_time = perf_counter()
                logging.debug(
                    f"generate TTFT time: {times[0]} s, {len(generated_token_ids)} tokens cost time: {sum(times)} s, avg cost time: {sum(times)/len(generated_token_ids)}"
                )
                self.assertGreater(len(generated_token_ids), 0)
                if "max_new_tokens" in case["kwargs"]:
                    max_new_tokens = case["kwargs"]["max_new_tokens"]
                    self.assertLessEqual(len(generated_token_ids), max_new_tokens)
                    if len(generated_token_ids) < max_new_tokens:
                        if "stop_ids" in case["kwargs"]:
                            for stop_id in case["kwargs"]["stop_ids"]:
                                self.assertIn(stop_id, generated_token_ids)
                generated_text = self.tokenizer.decode(generated_token_ids)
                logging.debug(f"generated text: {generated_text}")



import json
import locale
from pathlib import Path

I18N_FILE_PATH = Path(__file__).parent / "locale"
DEFAULT_LANGUAGE = "en_US"


def load_language_list(language):
    with open(I18N_FILE_PATH / f"{language}.json", "r", encoding="utf-8") as f:
        language_list = json.load(f)

    return language_list


class I18nAuto:
    def __init__(self):
        i18n_file = Path(".locale")

        if i18n_file.exists():
            with open(i18n_file, "r", encoding="utf-8") as f:
                language = f.read().strip()
        else:
            # getlocale can't identify the system's language ((None, None))
            language = locale.getdefaultlocale()[0]

        if (I18N_FILE_PATH / f"{language}.json").exists() is False:
            language = DEFAULT_LANGUAGE

        self.language = language
        self.language_map = load_language_list(language)

    def __call__(self, key):
        return self.language_map.get(key, key)

    def __repr__(self):
        return "Use Language: " + self.language


i18n = I18nAuto()



import os
import sys

sys.path.append(os.getcwd())

import argparse
import csv
import json
import shutil
from importlib.resources import files
from pathlib import Path

import torchaudio
from tqdm import tqdm
from datasets.arrow_writer import ArrowWriter

from f5_tts.model.utils import (
    convert_char_to_pinyin,
)


PRETRAINED_VOCAB_PATH = files("f5_tts").joinpath("../../data/Emilia_ZH_EN_pinyin/vocab.txt")


def is_csv_wavs_format(input_dataset_dir):
    fpath = Path(input_dataset_dir)
    metadata = fpath / "metadata.csv"
    wavs = fpath / "wavs"
    return metadata.exists() and metadata.is_file() and wavs.exists() and wavs.is_dir()


def prepare_csv_wavs_dir(input_dir):
    assert is_csv_wavs_format(input_dir), f"not csv_wavs format: {input_dir}"
    input_dir = Path(input_dir)
    metadata_path = input_dir / "metadata.csv"
    audio_path_text_pairs = read_audio_text_pairs(metadata_path.as_posix())

    sub_result, durations = [], []
    vocab_set = set()
    polyphone = True
    for audio_path, text in audio_path_text_pairs:
        if not Path(audio_path).exists():
            print(f"audio {audio_path} not found, skipping")
            continue
        audio_duration = get_audio_duration(audio_path)
        # assume tokenizer = "pinyin"  ("pinyin" | "char")
        text = convert_char_to_pinyin([text], polyphone=polyphone)[0]
        sub_result.append({"audio_path": audio_path, "text": text, "duration": audio_duration})
        durations.append(audio_duration)
        vocab_set.update(list(text))

    return sub_result, durations, vocab_set


def get_audio_duration(audio_path):
    audio, sample_rate = torchaudio.load(audio_path)
    return audio.shape[1] / sample_rate


def read_audio_text_pairs(csv_file_path):
    audio_text_pairs = []

    parent = Path(csv_file_path).parent
    with open(csv_file_path, mode="r", newline="", encoding="utf-8-sig") as csvfile:
        reader = csv.reader(csvfile, delimiter="|")
        next(reader)  # Skip the header row
        for row in reader:
            if len(row) >= 2:
                audio_file = row[0].strip()  # First column: audio file path
                text = row[1].strip()  # Second column: text
                audio_file_path = parent / audio_file
                audio_text_pairs.append((audio_file_path.as_posix(), text))

    return audio_text_pairs


def save_prepped_dataset(out_dir, result, duration_list, text_vocab_set, is_finetune):
    out_dir = Path(out_dir)
    # save preprocessed dataset to disk
    out_dir.mkdir(exist_ok=True, parents=True)
    print(f"\nSaving to {out_dir} ...")

    # dataset = Dataset.from_dict({"audio_path": audio_path_list, "text": text_list, "duration": duration_list})  # oom
    # dataset.save_to_disk(f"{out_dir}/raw", max_shard_size="2GB")
    raw_arrow_path = out_dir / "raw.arrow"
    with ArrowWriter(path=raw_arrow_path.as_posix(), writer_batch_size=1) as writer:
        for line in tqdm(result, desc="Writing to raw.arrow ..."):
            writer.write(line)

    # dup a json separately saving duration in case for DynamicBatchSampler ease
    dur_json_path = out_dir / "duration.json"
    with open(dur_json_path.as_posix(), "w", encoding="utf-8") as f:
        json.dump({"duration": duration_list}, f, ensure_ascii=False)

    # vocab map, i.e. tokenizer
    # add alphabets and symbols (optional, if plan to ft on de/fr etc.)
    # if tokenizer == "pinyin":
    #     text_vocab_set.update([chr(i) for i in range(32, 127)] + [chr(i) for i in range(192, 256)])
    voca_out_path = out_dir / "vocab.txt"
    with open(voca_out_path.as_posix(), "w") as f:
        for vocab in sorted(text_vocab_set):
            f.write(vocab + "\n")

    if is_finetune:
        file_vocab_finetune = PRETRAINED_VOCAB_PATH.as_posix()
        shutil.copy2(file_vocab_finetune, voca_out_path)
    else:
        with open(voca_out_path, "w") as f:
            for vocab in sorted(text_vocab_set):
                f.write(vocab + "\n")

    dataset_name = out_dir.stem
    print(f"\nFor {dataset_name}, sample count: {len(result)}")
    print(f"For {dataset_name}, vocab size is: {len(text_vocab_set)}")
    print(f"For {dataset_name}, total {sum(duration_list)/3600:.2f} hours")


def prepare_and_save_set(inp_dir, out_dir, is_finetune: bool = True):
    if is_finetune:
        assert PRETRAINED_VOCAB_PATH.exists(), f"pretrained vocab.txt not found: {PRETRAINED_VOCAB_PATH}"
    sub_result, durations, vocab_set = prepare_csv_wavs_dir(inp_dir)
    save_prepped_dataset(out_dir, sub_result, durations, vocab_set, is_finetune)


def cli():
    # finetune: python scripts/prepare_csv_wavs.py /path/to/input_dir /path/to/output_dir_pinyin
    # pretrain: python scripts/prepare_csv_wavs.py /path/to/output_dir_pinyin --pretrain
    parser = argparse.ArgumentParser(description="Prepare and save dataset.")
    parser.add_argument("inp_dir", type=str, help="Input directory containing the data.")
    parser.add_argument("out_dir", type=str, help="Output directory to save the prepared data.")
    parser.add_argument("--pretrain", action="store_true", help="Enable for new pretrain, otherwise is a fine-tune")

    args = parser.parse_args()

    prepare_and_save_set(args.inp_dir, args.out_dir, is_finetune=not args.pretrain)


if __name__ == "__main__":
    cli()



import os, sys

if sys.platform == "darwin":
    os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

now_dir = os.getcwd()
sys.path.append(now_dir)

import logging

import ChatTTS

from tools.logger import get_logger

logger = get_logger("Test #588", lv=logging.WARN)

chat = ChatTTS.Chat(logger)
chat.load(compile=False, source="huggingface")  # Set to True for better performance

texts = [
    "总结一下，AI Agent是大模型功能的扩展，让AI更接近于通用人工智能，也就是我们常说的AGI。",
    "你真是太聪明啦。",
]

fail = False

refined = chat.infer(
    texts,
    refine_text_only=True,
    stream=False,
    params_refine_text=ChatTTS.Chat.RefineTextParams(show_tqdm=False),
)

for i, t in enumerate(refined):
    if len(t) > 4 * len(texts[i]):
        fail = True
        logger.warning("in: %s, out: %s", texts[i], t)

if fail:
    import sys

    sys.exit(1)



from typing import Optional

import torch
from torch import nn
from torch.nn.utils import weight_norm

from decoder.modules import ConvNeXtBlock, ResBlock1, AdaLayerNorm


def nonlinearity(x):
    # swish
    return x * torch.sigmoid(x)


def Normalize(in_channels, num_groups=32):
    return torch.nn.GroupNorm(num_groups=num_groups, num_channels=in_channels, eps=1e-6, affine=True)


class ResnetBlock(nn.Module):
    def __init__(self, *, in_channels, out_channels=None, conv_shortcut=False,
                 dropout, temb_channels=512):
        super().__init__()
        self.in_channels = in_channels
        out_channels = in_channels if out_channels is None else out_channels
        self.out_channels = out_channels
        self.use_conv_shortcut = conv_shortcut

        self.norm1 = Normalize(in_channels)
        self.conv1 = torch.nn.Conv1d(in_channels,
                                     out_channels,
                                     kernel_size=3,
                                     stride=1,
                                     padding=1)
        if temb_channels > 0:
            self.temb_proj = torch.nn.Linear(temb_channels,
                                             out_channels)
        self.norm2 = Normalize(out_channels)
        self.dropout = torch.nn.Dropout(dropout)
        self.conv2 = torch.nn.Conv1d(out_channels,
                                     out_channels,
                                     kernel_size=3,
                                     stride=1,
                                     padding=1)
        if self.in_channels != self.out_channels:
            if self.use_conv_shortcut:
                self.conv_shortcut = torch.nn.Conv1d(in_channels,
                                                     out_channels,
                                                     kernel_size=3,
                                                     stride=1,
                                                     padding=1)
            else:
                self.nin_shortcut = torch.nn.Conv1d(in_channels,
                                                    out_channels,
                                                    kernel_size=1,
                                                    stride=1,
                                                    padding=0)

    def forward(self, x, temb=None):            
        h = x
        h = self.norm1(h)
        h = nonlinearity(h)
        h = self.conv1(h)

        if temb is not None:
            h = h + self.temb_proj(nonlinearity(temb))[:, :, None, None]

        h = self.norm2(h)
        h = nonlinearity(h)
        h = self.dropout(h)
        h = self.conv2(h)

        if self.in_channels != self.out_channels:
            if self.use_conv_shortcut:
                x = self.conv_shortcut(x)
            else:
                x = self.nin_shortcut(x)

        return x + h

class AttnBlock(nn.Module):
    def __init__(self, in_channels):
        super().__init__()
        self.in_channels = in_channels

        self.norm = Normalize(in_channels)
        self.q = torch.nn.Conv1d(in_channels,
                                 in_channels,
                                 kernel_size=1,
                                 stride=1,
                                 padding=0)
        self.k = torch.nn.Conv1d(in_channels,
                                 in_channels,
                                 kernel_size=1,
                                 stride=1,
                                 padding=0)
        self.v = torch.nn.Conv1d(in_channels,
                                 in_channels,
                                 kernel_size=1,
                                 stride=1,
                                 padding=0)
        self.proj_out = torch.nn.Conv1d(in_channels,
                                        in_channels,
                                        kernel_size=1,
                                        stride=1,
                                        padding=0)

    def forward(self, x):
        h_ = x
        h_ = self.norm(h_)
        q = self.q(h_)
        k = self.k(h_)
        v = self.v(h_)

        # compute attention
        b, c, h = q.shape
        q = q.permute(0, 2, 1)  # b,hw,c
        w_ = torch.bmm(q, k)  # b,hw,hw    w[b,i,j]=sum_c q[b,i,c]k[b,c,j]
        w_ = w_ * (int(c) ** (-0.5))
        w_ = torch.nn.functional.softmax(w_, dim=2)

        # attend to values
        w_ = w_.permute(0, 2, 1)  # b,hw,hw (first hw of k, second of q)
        h_ = torch.bmm(v, w_)  # b, c,hw (hw of q) h_[b,c,j] = sum_i v[b,c,i] w_[b,i,j]

        h_ = self.proj_out(h_)

        return x + h_

def make_attn(in_channels, attn_type="vanilla"):
    assert attn_type in ["vanilla", "linear", "none"], f'attn_type {attn_type} unknown'
    # print(f"making attention of type '{attn_type}' with {in_channels} in_channels")
    if attn_type == "vanilla":
        return AttnBlock(in_channels)


class Backbone(nn.Module):
    """Base class for the generator's backbone. It preserves the same temporal resolution across all layers."""

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        """
        Args:
            x (Tensor): Input tensor of shape (B, C, L), where B is the batch size,
                        C denotes output features, and L is the sequence length.

        Returns:
            Tensor: Output of shape (B, L, H), where B is the batch size, L is the sequence length,
                    and H denotes the model dimension.
        """
        raise NotImplementedError("Subclasses must implement the forward method.")


class VocosBackbone(Backbone):
    """
    Vocos backbone module built with ConvNeXt blocks. Supports additional conditioning with Adaptive Layer Normalization

    Args:
        input_channels (int): Number of input features channels.
        dim (int): Hidden dimension of the model.
        intermediate_dim (int): Intermediate dimension used in ConvNeXtBlock.
        num_layers (int): Number of ConvNeXtBlock layers.
        layer_scale_init_value (float, optional): Initial value for layer scaling. Defaults to `1 / num_layers`.
        adanorm_num_embeddings (int, optional): Number of embeddings for AdaLayerNorm.
                                                None means non-conditional model. Defaults to None.
    """

    def __init__(
        self,
        input_channels: int,
        dim: int,
        intermediate_dim: int,
        num_layers: int,
        layer_scale_init_value: Optional[float] = None,
        adanorm_num_embeddings: Optional[int] = None,
    ):
        super().__init__()
        self.input_channels = input_channels
        self.embed = nn.Conv1d(input_channels, dim, kernel_size=7, padding=3)
        self.adanorm = adanorm_num_embeddings is not None
        if adanorm_num_embeddings:
            self.norm = AdaLayerNorm(adanorm_num_embeddings, dim, eps=1e-6)
        else:
            self.norm = nn.LayerNorm(dim, eps=1e-6)
        layer_scale_init_value = layer_scale_init_value or 1 / num_layers
        self.convnext = nn.ModuleList(
            [
                ConvNeXtBlock(
                    dim=dim,
                    intermediate_dim=intermediate_dim,
                    layer_scale_init_value=layer_scale_init_value,
                    adanorm_num_embeddings=adanorm_num_embeddings,
                )
                for _ in range(num_layers)
            ]
        )
        self.final_layer_norm = nn.LayerNorm(dim, eps=1e-6)
        self.apply(self._init_weights)

        self.temb_ch = 0
        block_in = dim
        dropout = 0.1
        attn_type="vanilla"

        pos_net : tp.List[nn.Module] = [
            ResnetBlock(in_channels=block_in,out_channels=block_in,
                        temb_channels=self.temb_ch,dropout=dropout),
            ResnetBlock(in_channels=block_in,out_channels=block_in,
                        temb_channels=self.temb_ch,dropout=dropout),
            make_attn(block_in, attn_type=attn_type),
            ResnetBlock(in_channels=block_in,out_channels=block_in,
                    temb_channels=self.temb_ch,dropout=dropout),
            ResnetBlock(in_channels=block_in,out_channels=block_in,
                        temb_channels=self.temb_ch,dropout=dropout),
            Normalize(block_in)
        ]

        self.pos_net = nn.Sequential(*pos_net)

    def _init_weights(self, m):
        if isinstance(m, (nn.Conv1d, nn.Linear)):
            nn.init.trunc_normal_(m.weight, std=0.02)
            nn.init.constant_(m.bias, 0)

    def forward(self, x: torch.Tensor, bandwidth_id: Optional[torch.Tensor] = None) -> torch.Tensor:
        x = self.embed(x)
        x = self.pos_net(x)
        if self.adanorm:
            assert bandwidth_id is not None
            x = self.norm(x.transpose(1, 2), cond_embedding_id=bandwidth_id)
        else:
            x = self.norm(x.transpose(1, 2))
        x = x.transpose(1, 2)
        for conv_block in self.convnext:
            x = conv_block(x, cond_embedding_id=bandwidth_id)
        x = self.final_layer_norm(x.transpose(1, 2))
        return x


class VocosResNetBackbone(Backbone):
    """
    Vocos backbone module built with ResBlocks.

    Args:
        input_channels (int): Number of input features channels.
        dim (int): Hidden dimension of the model.
        num_blocks (int): Number of ResBlock1 blocks.
        layer_scale_init_value (float, optional): Initial value for layer scaling. Defaults to None.
    """

    def __init__(
        self, input_channels, dim, num_blocks, layer_scale_init_value=None,
    ):
        super().__init__()
        self.input_channels = input_channels
        self.embed = weight_norm(nn.Conv1d(input_channels, dim, kernel_size=3, padding=1))
        layer_scale_init_value = layer_scale_init_value or 1 / num_blocks / 3
        self.resnet = nn.Sequential(
            *[ResBlock1(dim=dim, layer_scale_init_value=layer_scale_init_value) for _ in range(num_blocks)]
        )

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        x = self.embed(x)
        x = self.resnet(x)
        x = x.transpose(1, 2)
        return x



#
# Copyright (c) 2024, Daily
#
# SPDX-License-Identifier: BSD 2-Clause License
#

import sys

from enum import Enum

if sys.version_info < (3, 11):

    class StrEnum(str, Enum):
        def __new__(cls, value):
            obj = str.__new__(cls, value)
            obj._value_ = value
            return obj
else:
    from enum import StrEnum


class Language(StrEnum):
    BG = "bg"  # Bulgarian
    CA = "ca"  # Catalan
    ZH = "zh"  # Chinese simplified
    ZH_CN = "zh-CN"  # Chinese simplified
    ZH_TW = "zh-TW"  # Chinese traditional
    ZN = "zn"  # Chinese and English
    CS = "cs"  # Czech
    DA = "da"  # Danish
    NL = "nl"  # Dutch
    EN = "en"  # English
    EN_US = "en-US"  # English (USA)
    EN_AU = "en-AU"  # English (Australia)
    EN_GB = "en-GB"  # English (Great Britain)
    EN_NZ = "en-NZ"  # English (New Zealand)
    EN_IN = "en-IN"  # English (India)
    ET = "et"  # Estonian
    FI = "fi"  # Finnish
    NL_BE = "nl-BE"  # Flemmish
    FR = "fr"  # French
    FR_CA = "fr-CA"  # French (Canada)
    DE = "de"  # German
    DE_CH = "de-CH"  # German (Switzerland)
    EL = "el"  # Greek
    HI = "hi"  # Hindi
    HU = "hu"  # Hungarian
    ID = "id"  # Indonesian
    IT = "it"  # Italian
    JA = "ja"  # Japanese
    KO = "ko"  # Korean
    LV = "lv"  # Latvian
    LT = "lt"  # Lithuanian
    MS = "ms"  # Malay
    NO = "no"  # Norwegian
    PL = "pl"  # Polish
    PT = "pt"  # Portuguese
    PT_BR = "pt-BR"  # Portuguese (Brazil)
    RO = "ro"  # Romanian
    RU = "ru"  # Russian
    SK = "sk"  # Slovak
    ES = "es"  # Spanish
    SV = "sv"  # Swedish
    TH = "th"  # Thai
    TR = "tr"  # Turkish
    UK = "uk"  # Ukrainian
    VI = "vi"  # Vietnamese


TO_LLM_LANGUAGE = {
    "bg": "Bulgarian",
    "ca": "Catalan",
    "zh": "Chinese simplified(简体中文)",
    "zh-CN": "Chinese simplified(简体中文)",
    "zh-TW": "Chinese traditional(繁体中文)",
    "zn": "Chinese and English",
    "cs": "Czech",
    "da": "Danish",
    "nl": "Dutch",
    "en": "English",
    "en-US": "English (USA)",
    "en-AU": "English (Australia)",
    "en-GB": "English (Great Britain)",
    "en-NZ": "English (New Zealand)",
    "en-IN": "English (India)",
    "et": "Estonian",
    "fi": "Finnish",
    "nl-BE": "Flemmish",
    "fr": "French",
    "fr-CA": "French (Canada)",
    "de": "German",
    "de-CH": "German (Switzerland)",
    "el": "Greek",
    "hi": "Hindi",
    "hu": "Hungarian",
    "id": "Indonesian",
    "it": "Italian",
    "ja": "Japanese",
    "ko": "Korean",
    "lv": "Latvian",
    "lt": "Lithuanian",
    "ms": "Malay",
    "no": "Norwegian",
    "pl": "Polish",
    "pt": "Portuguese",
    "pt-BR": "Portuguese (Brazil)",
    "ro": "Romanian",
    "ru": "Russian",
    "sk": "Slovak",
    "es": "Spanish",
    "sv": "Swedish",
    "th": "Thai",
    "tr": "Turkish",
    "uk": "Ukrainian",
    "vi": "Vietnamese",
}



import asyncio
from dataclasses import dataclass

import aiohttp
from apipeline.frames.sys_frames import Frame
from apipeline.pipeline.pipeline import Pipeline, FrameProcessor
from apipeline.pipeline.sync_parallel_pipeline import SyncParallelPipeline
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.pipeline.runner import PipelineRunner
from apipeline.frames.data_frames import TextFrame, ImageRawFrame, AudioRawFrame
from apipeline.processors.frame_processor import FrameDirection
from apipeline.processors.aggregators.sentence import SentenceAggregator
from apipeline.processors.logger import FrameLogger

from src.processors.translation.google_translate_processor import GoogleTranslateProcessor
from src.processors.speech.tts.tts_processor import TTSProcessor
from src.cmd.bots.base_daily import DailyRoomBot
from src.common.types import DailyParams
from src.types.frames.control_frames import LLMFullResponseStartFrame
from src.transports.daily import DailyTransport
from src.types.frames.data_frames import LLMMessagesFrame
from .. import register_ai_room_bots


@dataclass
class MonthFrame(Frame):
    month: str

    def __str__(self):
        return f"{self.name}(month: {self.month})"


class MonthPrepender(FrameProcessor):
    def __init__(self):
        super().__init__()
        self.most_recent_month = ""
        self.prepend_to_next_text_frame = False

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, MonthFrame):
            self.most_recent_month = frame.month
        elif self.prepend_to_next_text_frame and isinstance(frame, TextFrame):
            await self.push_frame(TextFrame(f"{self.most_recent_month}: {frame.text}"))
            self.prepend_to_next_text_frame = False
        elif isinstance(frame, LLMFullResponseStartFrame):
            self.prepend_to_next_text_frame = True
            await self.push_frame(frame)
        else:
            await self.push_frame(frame, direction)


@register_ai_room_bots.register
class DailyMonthNarrationBot(DailyRoomBot):
    def __init__(self, **args) -> None:
        super().__init__(**args)
        self.init_bot_config()

    async def arun(self):
        async with aiohttp.ClientSession() as session:
            daily_params = DailyParams(
                audio_out_enabled=True,
                camera_out_enabled=True,
                camera_out_width=1280,
                camera_out_height=720,
            )

            tts_processor: TTSProcessor = self.get_tts_processor()
            stream_info = tts_processor.get_stream_info()
            daily_params.audio_out_sample_rate = stream_info["sample_rate"]
            daily_params.audio_out_channels = stream_info["channels"]

            transport = DailyTransport(
                self.args.room_url,
                self.args.token,
                self.args.bot_name,
                daily_params,
            )
            transport.add_event_handler(
                "on_first_participant_joined", self.on_first_participant_joined
            )
            transport.add_event_handler("on_participant_left", self.on_participant_left)
            transport.add_event_handler("on_call_state_updated", self.on_call_state_updated)

            llm_processor = self.get_llm_processor()
            sentence_aggregator = SentenceAggregator()
            month_prepender = MonthPrepender()

            image_gen_processor = self.get_image_gen_processor()
            image_gen_processor.set_aiohttp_session(session)
            image_gen_processor.set_size(
                width=daily_params.camera_out_width,
                height=daily_params.camera_out_height,
            )

            translate_processor = GoogleTranslateProcessor()

            frames = []
            for month in [
                "January",
                "February",
                "March",
                "April",
                "May",
                "June",
                "July",
                "August",
                "September",
                "October",
                "November",
                "December",
            ]:
                messages = [
                    {
                        "role": "system",
                        "content": f"Describe a nature photograph suitable for use in a calendar, for the month of {month}. Include only the image description with no preamble. Limit the description to one sentence, please.",
                    }
                ]
                frames.append(MonthFrame(month=month))
                frames.append(LLMMessagesFrame(messages))

            self.task = PipelineTask(
                Pipeline(
                    [
                        llm_processor,  # LLM
                        sentence_aggregator,  # Aggregates LLM output into full sentences
                        SyncParallelPipeline(  # Run pipelines in parallel aggregating the result
                            # Create "Month: sentence" and output audio
                            [month_prepender, translate_processor, tts_processor],
                            [image_gen_processor],  # Generate image
                        ),
                        FrameLogger(include_frame_types=[ImageRawFrame, AudioRawFrame]),
                        transport.output_processor(),
                    ]
                )
            )
            await self.task.queue_frames(frames)
            await PipelineRunner().run(self.task)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# flake8: noqa

"""EnCodec neural audio codec."""

__version__ = "0.1.2a3"

from .model import EncodecModel



import glob
import logging
import os
from typing import List, Union, Tuple

import numpy as np
import torch

from . import Code2WavEngineConfig, Code2WavGenerationConfig

from ..model_loader.weight_utils import safetensors_weights_iterator
from .modeling_fast import Qwen2Code2wav


class Code2WavEngine:
    def __init__(self, **kwargs) -> None:
        self.args = Code2WavEngineConfig(**kwargs)
        model_path = self.args.model_path
        enable_torch_compile = self.args.enable_torch_compile
        enable_torch_compile_first_chunk = self.args.enable_torch_compile_first_chunk
        odeint_method = self.args.odeint_method
        odeint_method_relaxed = self.args.odeint_method_relaxed
        batched_chunk = self.args.batched_chunk
        frequency: str = self.args.frequency
        device: Union[int, str] = self.args.device
        code2wav_dynamic_batch: bool = self.args.code2wav_dynamic_batch  # todo batch chunk

        if isinstance(device, int):
            device = f"cuda:{device}"
        self.device = torch.device(device)

        logging.info(
            f"Code2WavEngine starting up on device {self.device}, with model {model_path}, method: {odeint_method}, relaxed: {odeint_method_relaxed}"
        )

        # load spk_dict ["Ethan", "Chelsie"]
        if os.path.exists(os.path.join(model_path, "spk_dict.pt")):
            self.code2wav_conds, self.code2wav_ref_mels = self.load_spk_dict(model_path)
        assert len(self.code2wav_conds) > 0 and len(self.code2wav_ref_mels) > 0, "No speakers found"
        if "default" not in self.code2wav_conds:
            self.code2wav_conds["default"] = self.code2wav_conds[
                list(self.code2wav_conds.keys())[0]
            ]
        if "default" not in self.code2wav_ref_mels:
            self.code2wav_ref_mels["default"] = self.code2wav_ref_mels[
                list(self.code2wav_ref_mels.keys())[0]
            ]

        self.frequency = frequency
        self.code2wav_steps: int = 10
        self.code2wav_bs_mel: int = 24 if frequency == "50hz" else 32
        self.factor: int = 2 if frequency == "50hz" else 4

        dit_model, bigvgan_model = self.load_code2wav(model_path)
        self.code2wav = Qwen2Code2wav(
            dit_ckpt=dit_model,
            bigvgan_ckpt=bigvgan_model,
            steps=self.code2wav_steps,
            bs_mel=self.code2wav_bs_mel,
            odeint_method=odeint_method,
            odeint_method_relaxed=odeint_method_relaxed,
            batched_chunk=batched_chunk,
            frequency=frequency,
            device=self.device,
            with_weight_norm=False,
        )
        self.torch_compile_model(enable_torch_compile, enable_torch_compile_first_chunk)

        self.code2wav_y_all = torch.randn(
            1, 32768, 80, device=self.device, dtype=list(self.code2wav_ref_mels.values())[0].dtype
        )

    def get_voice(self, voice_type: str = "default"):
        if voice_type not in self.code2wav_conds:
            logging.warning(f"voice type {voice_type} not found, using default")
            voice_type = "default"
        code2wav_cond = self.code2wav_conds[voice_type]
        code2wav_ref_mel = self.code2wav_ref_mels[voice_type]
        return code2wav_cond, code2wav_ref_mel

    def load_spk_dict(self, model_path):
        code2wav_conds, code2wav_ref_mels = {}, {}

        if not os.path.exists(os.path.join(model_path, "spk_dict.pt")):
            return code2wav_conds, code2wav_ref_mels

        for key, value in torch.load(os.path.join(model_path, "spk_dict.pt")).items():
            code2wav_conds[key] = value["cond"].to(self.device)
            code2wav_ref_mels[key] = value["ref_mel"].to(self.device)
        return code2wav_conds, code2wav_ref_mels

    def load_code2wav(self, model_path):
        dit_model, bigvgan_model = {}, {}
        safetensors = sorted(glob.glob(os.path.join(model_path, "*.safetensors")))
        legacy_weights = False
        for key, value in safetensors_weights_iterator(safetensors, use_tqdm_on_load=True):
            legacy_weights = legacy_weights or "input_embed.spk_encoder.fc.conv.weight" in key
            if legacy_weights:
                break
        for key, value in safetensors_weights_iterator(safetensors, use_tqdm_on_load=True):
            if key.startswith("token2wav.code2wav_bigvgan_model."):
                if "generator" not in bigvgan_model:
                    bigvgan_model["generator"] = {}
                bigvgan_model["generator"][key.replace("token2wav.code2wav_bigvgan_model.", "")] = (
                    value
                )
            if key.startswith("token2wav.code2wav_dit_model."):
                key = key.replace("token2wav.code2wav_dit_model.", "transformer.")
                if key.startswith("transformer.input_embed.spk_encoder"):
                    if legacy_weights:
                        dit_model[key] = value
                    else:
                        dit_model[
                            key.replace(".bias", ".conv.bias").replace(".weight", ".conv.weight")
                        ] = value
                elif ".ff.ff.0.weight" in key or ".ff.ff.0.bias" in key:
                    dit_model[
                        key.replace(".ff.ff.0.weight", ".ff.ff.0.0.weight").replace(
                            ".ff.ff.0.bias", ".ff.ff.0.0.bias"
                        )
                    ] = value
                elif ".ff.ff.3.weight" in key or ".ff.ff.3.bias" in key:
                    dit_model[
                        key.replace(".ff.ff.3.weight", ".ff.ff.2.weight").replace(
                            ".ff.ff.3.bias", ".ff.ff.2.bias"
                        )
                    ] = value
                else:
                    dit_model[key] = value
        return dit_model, bigvgan_model

    def torch_compile_model(
        self,
        enable_torch_compile,
        enable_torch_compile_first_chunk,
    ):
        if not enable_torch_compile:
            return

        # compile the bigvgan
        self.code2wav.code2wav_bigvgan_model.vocoder.forward = torch.compile(
            self.code2wav.code2wav_bigvgan_model.vocoder.forward,
        )
        # compile the dit
        if hasattr(self.code2wav, "enable_torch_compile"):
            self.code2wav.enable_torch_compile(enable_torch_compile_first_chunk)

        logging.info("Code2Wav model torch compiled")

    @torch.inference_mode()
    def step_generate_waveform(
        self,
        code: List[int],
        prev_generated: Union[torch.Tensor, List[torch.Tensor]],
        progress: int,
        finished: bool = False,
        y_all: torch.Tensor = None,
        voice_type: str = "default",
        gen_args: Code2WavGenerationConfig = Code2WavGenerationConfig(),
    ) -> Tuple[Union[torch.Tensor, List[torch.Tensor]], torch.Tensor]:
        """
        Generate waveform from code list step by step.
        """
        cond, ref_mel = self.get_voice(voice_type)
        chunk_code_length = len(code) * self.factor - self.code2wav.future_cache_size
        if (
            chunk_code_length > 0 and chunk_code_length % self.code2wav.chunk_size == 0
        ) or finished:
            code = torch.tensor(code, dtype=torch.long, device=self.device).reshape(1, -1)
            if progress == 0 and finished:
                process_chunk = self.code2wav.process_little_chunk
            else:
                process_chunk = self.code2wav.process_chunk

            return process_chunk(
                cond,
                ref_mel,
                codec_all=code,
                y_all=self.code2wav_y_all if y_all is None else y_all,
                i=progress,
                steps=gen_args.num_steps,
                prev_generated=prev_generated,
                finished=finished,
                cfg_strength=gen_args.guidance_scale,
                sway_sampling_coef=gen_args.sway_coefficient,
            )
        else:
            return prev_generated, None



from dataclasses import dataclass, field
from typing import Any, List


from apipeline.frames.data_frames import Frame, DataFrame, TextFrame, ImageRawFrame, AudioRawFrame


@dataclass
class URLImageRawFrame(ImageRawFrame):
    """An image with an associated URL. Will be shown by the transport if the
    transport's camera is enabled.

    """

    url: str | None

    def __str__(self):
        return f"{self.name}(url: {self.url}, size: {self.size}, format: {self.format}), mode:{self.mode}"


@dataclass
class VisionImageRawFrame(ImageRawFrame):
    """An image with an associated text to ask for a description of it. Will be
    shown by the transport if the transport's camera is enabled.

    """

    text: str | None

    def __str__(self):
        return f"{self.name}(text: {self.text}, size: {self.size}, format: {self.format}, bytes_len:{len(self.image)}, mode:{self.mode}"


@dataclass
class UserImageRawFrame(ImageRawFrame):
    """An image associated to a user. Will be shown by the transport if the
    transport's camera is enabled.

    """

    user_id: str

    def __str__(self):
        return f"{self.name}(user: {self.user_id}, size: {self.size}, format: {self.format}), mode:{self.mode}"


@dataclass
class SpriteFrame(Frame):
    """An animated sprite. Will be shown by the transport if the transport's
    camera is enabled. Will play at the framerate specified in the transport's
    `fps` constructor parameter.

    """

    images: List[ImageRawFrame]

    def __str__(self):
        return f"{self.name}(size: {len(self.images)})"


@dataclass
class TranscriptionFrame(TextFrame):
    """A text frame with transcription-specific data. Will be placed in the
    transport's receive queue when a participant speaks.

    """

    user_id: str
    timestamp: str
    language: str | None = None

    def __str__(self):
        return f"{self.name}(user: {self.user_id}, text: {self.text}, timestamp: {self.timestamp}, language: {self.language})"


@dataclass
class InterimTranscriptionFrame(TextFrame):
    """A text frame with interim transcription-specific data. Will be placed in
    the transport's receive queue when a participant speaks."""

    user_id: str
    timestamp: str
    language: str

    def __str__(self):
        return f"{self.name}(user: {self.user_id}, text: {self.text}, timestamp: {self.timestamp}, language: {self.language})"


@dataclass
class LLMMessagesFrame(DataFrame):
    """A frame containing a list of LLM messages. Used to signal that an LLM
    service should run a chat completion and emit an LLMStartFrames, TextFrames
    and an LLMEndFrame. Note that the messages property on this class is
    mutable, and will be be updated by various ResponseAggregator frame
    processors.

    """

    messages: List[dict]

    def __str__(self):
        return f"{self.name}(messages: {self.messages})"


@dataclass
class TransportMessageFrame(DataFrame):
    message: Any
    urgent: bool = False

    def __str__(self):
        return f"{self.name}(message: {self.message})"


@dataclass
class LLMMessagesAppendFrame(DataFrame):
    """A frame containing a list of LLM messages that neeed to be added to the
    current context.

    """

    messages: List[dict]


@dataclass
class LLMMessagesUpdateFrame(DataFrame):
    """A frame containing a list of new LLM messages. These messages will
    replace the current context LLM messages and should generate a new
    LLMMessagesFrame.

    """

    messages: List[dict]


@dataclass
class LLMSetToolsFrame(DataFrame):
    """A frame containing a list of tools for an LLM to use for function calling.
    The specific format depends on the LLM being used, but it should typically
    contain JSON Schema objects.
    """

    tools: List[dict]


@dataclass
class TTSSpeakFrame(DataFrame):
    """A frame that contains a text that should be spoken by the TTS in the
    pipeline (if any).

    """

    text: str


@dataclass
class DailyTransportMessageFrame(TransportMessageFrame):
    participant_id: str | None = None


@dataclass
class LivekitTransportMessageFrame(TransportMessageFrame):
    participant_id: str | None = None


@dataclass
class AgoraTransportMessageFrame(TransportMessageFrame):
    participant_id: str | None = None


@dataclass
class FunctionCallResultFrame(DataFrame):
    """A frame containing the result of an LLM function (tool) call."""

    function_name: str
    tool_call_id: str
    arguments: dict
    result: Any


@dataclass
class PathAudioRawFrame(AudioRawFrame):
    """An audio with saved path."""

    path: str = ""

    def __post_init__(self):
        super().__post_init__()

    def __str__(self):
        return f"path:{self.path} {super().__str__()}"


@dataclass
class UserAudioRawFrame(AudioRawFrame):
    """An audio associated to a user. Will be shown by the transport if the
    transport's audio is enabled.

    """

    user_id: str = ""

    def __post_init__(self):
        super().__post_init__()

    def __str__(self):
        return f"user_id:{self.user_id} {super().__str__()}"


@dataclass
class UserVoiceRawFrame(UserAudioRawFrame):
    """
    An user audio associated to llm response text.
    """

    text: str = ""

    def __post_init__(self):
        super().__post_init__()

    def __str__(self):
        return f"text:{self.text} {super().__str__()}"


@dataclass
class VisionImageVoiceRawFrame(DataFrame):
    """An image + audio with an instruct text to ask for a description of it. Will be
    shown by the transport if the transport's camera is enabled.

    """

    text: str | None = None
    audio: AudioRawFrame | None = None
    images: List[ImageRawFrame] = field(default_factory=list)

    def __str__(self):
        s = f"{self.name}(text: {self.text}, audio:{self.audio}, images:"
        for image in self.images:
            s += f"{image}, "
        s += ")"
        return s


@dataclass
class UserVisionImageVoiceRawFrame(VisionImageVoiceRawFrame):
    """An user image + audio with an instruct text to ask for a description of it. Will be
    shown by the transport if the transport's camera is enabled.

    """

    user_id: str = ""

    def __str__(self):
        return f"user_id:{self.user_id} {super().__str__()}"



import asyncio
import logging

from apipeline.frames.sys_frames import Frame
from apipeline.frames.data_frames import TextFrame
from apipeline.pipeline.pipeline import Pipeline, FrameProcessor, FrameDirection
from deep_translator import GoogleTranslator


class GoogleTranslateProcessor(FrameProcessor):
    def __init__(self, src: str = "en", target: str = "zh-CN", is_keep_original=False):
        super().__init__()
        self._src = src
        self._target = target
        self._is_keep_original = is_keep_original
        self.translator = GoogleTranslator(source=src, target=target)
        self._translate_frame = TextFrame

    def set_translate_frame(self, frame: TextFrame):
        self._translate_frame = frame

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)
        if isinstance(frame, self._translate_frame):
            if self._src == self._target:
                await self.push_frame(frame, direction)
                return
            try_cn = 3
            while try_cn > 0:
                try:
                    translated_text = self.translator.translate(frame.text)
                    break
                except Exception as e:
                    logging.error("An error occurred:", e)
                    await asyncio.sleep(1)
                    try_cn -= 1
            if self._is_keep_original:
                translated_text = f"{frame.text}\n{translated_text}"
            await self.push_frame(self._translate_frame(text=translated_text))
        else:
            await self.push_frame(frame, direction)



import torch
import torchaudio
import gradio as gr
from os import getenv

from zonos.model import Zonos, DEFAULT_BACKBONE_CLS as ZonosBackbone
from zonos.conditioning import make_cond_dict, supported_language_codes
from zonos.utils import DEFAULT_DEVICE as device

CURRENT_MODEL_TYPE = None
CURRENT_MODEL = None

SPEAKER_EMBEDDING = None
SPEAKER_AUDIO_PATH = None


def load_model_if_needed(model_choice: str):
    global CURRENT_MODEL_TYPE, CURRENT_MODEL
    if CURRENT_MODEL_TYPE != model_choice:
        if CURRENT_MODEL is not None:
            del CURRENT_MODEL
            torch.cuda.empty_cache()
        print(f"Loading {model_choice} model...")
        CURRENT_MODEL = Zonos.from_pretrained(model_choice, device=device)
        CURRENT_MODEL.requires_grad_(False).eval()
        CURRENT_MODEL_TYPE = model_choice
        print(f"{model_choice} model loaded successfully!")
    return CURRENT_MODEL


def update_ui(model_choice):
    """
    Dynamically show/hide UI elements based on the model's conditioners.
    We do NOT display 'language_id' or 'ctc_loss' even if they exist in the model.
    """
    model = load_model_if_needed(model_choice)
    cond_names = [c.name for c in model.prefix_conditioner.conditioners]
    print("Conditioners in this model:", cond_names)

    text_update = gr.update(visible=("espeak" in cond_names))
    language_update = gr.update(visible=("espeak" in cond_names))
    speaker_audio_update = gr.update(visible=("speaker" in cond_names))
    prefix_audio_update = gr.update(visible=True)
    emotion1_update = gr.update(visible=("emotion" in cond_names))
    emotion2_update = gr.update(visible=("emotion" in cond_names))
    emotion3_update = gr.update(visible=("emotion" in cond_names))
    emotion4_update = gr.update(visible=("emotion" in cond_names))
    emotion5_update = gr.update(visible=("emotion" in cond_names))
    emotion6_update = gr.update(visible=("emotion" in cond_names))
    emotion7_update = gr.update(visible=("emotion" in cond_names))
    emotion8_update = gr.update(visible=("emotion" in cond_names))
    vq_single_slider_update = gr.update(visible=("vqscore_8" in cond_names))
    fmax_slider_update = gr.update(visible=("fmax" in cond_names))
    pitch_std_slider_update = gr.update(visible=("pitch_std" in cond_names))
    speaking_rate_slider_update = gr.update(visible=("speaking_rate" in cond_names))
    dnsmos_slider_update = gr.update(visible=("dnsmos_ovrl" in cond_names))
    speaker_noised_checkbox_update = gr.update(visible=("speaker_noised" in cond_names))
    unconditional_keys_update = gr.update(
        choices=[name for name in cond_names if name not in ("espeak", "language_id")]
    )

    return (
        text_update,
        language_update,
        speaker_audio_update,
        prefix_audio_update,
        emotion1_update,
        emotion2_update,
        emotion3_update,
        emotion4_update,
        emotion5_update,
        emotion6_update,
        emotion7_update,
        emotion8_update,
        vq_single_slider_update,
        fmax_slider_update,
        pitch_std_slider_update,
        speaking_rate_slider_update,
        dnsmos_slider_update,
        speaker_noised_checkbox_update,
        unconditional_keys_update,
    )


def generate_audio(
    model_choice,
    text,
    language,
    speaker_audio,
    prefix_audio,
    e1,
    e2,
    e3,
    e4,
    e5,
    e6,
    e7,
    e8,
    vq_single,
    fmax,
    pitch_std,
    speaking_rate,
    dnsmos_ovrl,
    speaker_noised,
    cfg_scale,
    min_p,
    seed,
    randomize_seed,
    unconditional_keys,
    progress=gr.Progress(),
):
    """
    Generates audio based on the provided UI parameters.
    We do NOT use language_id or ctc_loss even if the model has them.
    """
    selected_model = load_model_if_needed(model_choice)

    speaker_noised_bool = bool(speaker_noised)
    fmax = float(fmax)
    pitch_std = float(pitch_std)
    speaking_rate = float(speaking_rate)
    dnsmos_ovrl = float(dnsmos_ovrl)
    cfg_scale = float(cfg_scale)
    min_p = float(min_p)
    seed = int(seed)
    max_new_tokens = 86 * 30

    # This is a bit ew, but works for now.
    global SPEAKER_AUDIO_PATH, SPEAKER_EMBEDDING

    if randomize_seed:
        seed = torch.randint(0, 2**32 - 1, (1,)).item()
    torch.manual_seed(seed)

    if speaker_audio is not None and "speaker" not in unconditional_keys:
        if speaker_audio != SPEAKER_AUDIO_PATH:
            print("Recomputed speaker embedding")
            wav, sr = torchaudio.load(speaker_audio)
            SPEAKER_EMBEDDING = selected_model.make_speaker_embedding(wav, sr)
            SPEAKER_EMBEDDING = SPEAKER_EMBEDDING.to(device, dtype=torch.bfloat16)
            SPEAKER_AUDIO_PATH = speaker_audio

    audio_prefix_codes = None
    if prefix_audio is not None:
        wav_prefix, sr_prefix = torchaudio.load(prefix_audio)
        wav_prefix = wav_prefix.mean(0, keepdim=True)
        wav_prefix = selected_model.autoencoder.preprocess(wav_prefix, sr_prefix)
        wav_prefix = wav_prefix.to(device, dtype=torch.float32)
        audio_prefix_codes = selected_model.autoencoder.encode(wav_prefix.unsqueeze(0))

    emotion_tensor = torch.tensor(list(map(float, [e1, e2, e3, e4, e5, e6, e7, e8])), device=device)

    vq_val = float(vq_single)
    vq_tensor = torch.tensor([vq_val] * 8, device=device).unsqueeze(0)

    cond_dict = make_cond_dict(
        text=text,
        language=language,
        speaker=SPEAKER_EMBEDDING,
        emotion=emotion_tensor,
        vqscore_8=vq_tensor,
        fmax=fmax,
        pitch_std=pitch_std,
        speaking_rate=speaking_rate,
        dnsmos_ovrl=dnsmos_ovrl,
        speaker_noised=speaker_noised_bool,
        device=device,
        unconditional_keys=unconditional_keys,
    )
    conditioning = selected_model.prepare_conditioning(cond_dict)

    estimated_generation_duration = 30 * len(text) / 400
    estimated_total_steps = int(estimated_generation_duration * 86)

    def update_progress(_frame: torch.Tensor, step: int, _total_steps: int) -> bool:
        progress((step, estimated_total_steps))
        return True

    codes = selected_model.generate(
        prefix_conditioning=conditioning,
        audio_prefix_codes=audio_prefix_codes,
        max_new_tokens=max_new_tokens,
        cfg_scale=cfg_scale,
        batch_size=1,
        sampling_params=dict(min_p=min_p),
        callback=update_progress,
    )

    wav_out = selected_model.autoencoder.decode(codes).cpu().detach()
    sr_out = selected_model.autoencoder.sampling_rate
    if wav_out.dim() == 2 and wav_out.size(0) > 1:
        wav_out = wav_out[0:1, :]
    return (sr_out, wav_out.squeeze().numpy()), seed


def build_interface():
    supported_models = []
    if "transformer" in ZonosBackbone.supported_architectures:
        supported_models.append("Zyphra/Zonos-v0.1-transformer")

    if "hybrid" in ZonosBackbone.supported_architectures:
        supported_models.append("Zyphra/Zonos-v0.1-hybrid")
    else:
        print(
            "| The current ZonosBackbone does not support the hybrid architecture, meaning only the transformer model will be available in the model selector.\n"
            "| This probably means the mamba-ssm library has not been installed."
        )

    with gr.Blocks() as demo:
        with gr.Row():
            with gr.Column():
                model_choice = gr.Dropdown(
                    choices=supported_models,
                    value=supported_models[0],
                    label="Zonos Model Type",
                    info="Select the model variant to use.",
                )
                text = gr.Textbox(
                    label="Text to Synthesize",
                    value="Zonos uses eSpeak for text to phoneme conversion!",
                    lines=4,
                    max_length=500,  # approximately
                )
                language = gr.Dropdown(
                    choices=supported_language_codes,
                    value="en-us",
                    label="Language Code",
                    info="Select a language code.",
                )
            prefix_audio = gr.Audio(
                value="assets/silence_100ms.wav",
                label="Optional Prefix Audio (continue from this audio)",
                type="filepath",
            )
            with gr.Column():
                speaker_audio = gr.Audio(
                    label="Optional Speaker Audio (for cloning)",
                    type="filepath",
                )
                speaker_noised_checkbox = gr.Checkbox(label="Denoise Speaker?", value=False)

        with gr.Row():
            with gr.Column():
                gr.Markdown("## Conditioning Parameters")
                dnsmos_slider = gr.Slider(1.0, 5.0, value=4.0, step=0.1, label="DNSMOS Overall")
                fmax_slider = gr.Slider(0, 24000, value=24000, step=1, label="Fmax (Hz)")
                vq_single_slider = gr.Slider(0.5, 0.8, 0.78, 0.01, label="VQ Score")
                pitch_std_slider = gr.Slider(0.0, 300.0, value=45.0, step=1, label="Pitch Std")
                speaking_rate_slider = gr.Slider(5.0, 30.0, value=15.0, step=0.5, label="Speaking Rate")

            with gr.Column():
                gr.Markdown("## Generation Parameters")
                cfg_scale_slider = gr.Slider(1.0, 5.0, 2.0, 0.1, label="CFG Scale")
                min_p_slider = gr.Slider(0.0, 1.0, 0.15, 0.01, label="Min P")
                seed_number = gr.Number(label="Seed", value=420, precision=0)
                randomize_seed_toggle = gr.Checkbox(label="Randomize Seed (before generation)", value=True)

        with gr.Accordion("Advanced Parameters", open=False):
            gr.Markdown(
                "### Unconditional Toggles\n"
                "Checking a box will make the model ignore the corresponding conditioning value and make it unconditional.\n"
                'Practically this means the given conditioning feature will be unconstrained and "filled in automatically".'
            )
            with gr.Row():
                unconditional_keys = gr.CheckboxGroup(
                    [
                        "speaker",
                        "emotion",
                        "vqscore_8",
                        "fmax",
                        "pitch_std",
                        "speaking_rate",
                        "dnsmos_ovrl",
                        "speaker_noised",
                    ],
                    value=["emotion"],
                    label="Unconditional Keys",
                )

            gr.Markdown(
                "### Emotion Sliders\n"
                "Warning: The way these sliders work is not intuitive and may require some trial and error to get the desired effect.\n"
                "Certain configurations can cause the model to become unstable. Setting emotion to unconditional may help."
            )
            with gr.Row():
                emotion1 = gr.Slider(0.0, 1.0, 1.0, 0.05, label="Happiness")
                emotion2 = gr.Slider(0.0, 1.0, 0.05, 0.05, label="Sadness")
                emotion3 = gr.Slider(0.0, 1.0, 0.05, 0.05, label="Disgust")
                emotion4 = gr.Slider(0.0, 1.0, 0.05, 0.05, label="Fear")
            with gr.Row():
                emotion5 = gr.Slider(0.0, 1.0, 0.05, 0.05, label="Surprise")
                emotion6 = gr.Slider(0.0, 1.0, 0.05, 0.05, label="Anger")
                emotion7 = gr.Slider(0.0, 1.0, 0.1, 0.05, label="Other")
                emotion8 = gr.Slider(0.0, 1.0, 0.2, 0.05, label="Neutral")

        with gr.Column():
            generate_button = gr.Button("Generate Audio")
            output_audio = gr.Audio(label="Generated Audio", type="numpy", autoplay=True)

        model_choice.change(
            fn=update_ui,
            inputs=[model_choice],
            outputs=[
                text,
                language,
                speaker_audio,
                prefix_audio,
                emotion1,
                emotion2,
                emotion3,
                emotion4,
                emotion5,
                emotion6,
                emotion7,
                emotion8,
                vq_single_slider,
                fmax_slider,
                pitch_std_slider,
                speaking_rate_slider,
                dnsmos_slider,
                speaker_noised_checkbox,
                unconditional_keys,
            ],
        )

        # On page load, trigger the same UI refresh
        demo.load(
            fn=update_ui,
            inputs=[model_choice],
            outputs=[
                text,
                language,
                speaker_audio,
                prefix_audio,
                emotion1,
                emotion2,
                emotion3,
                emotion4,
                emotion5,
                emotion6,
                emotion7,
                emotion8,
                vq_single_slider,
                fmax_slider,
                pitch_std_slider,
                speaking_rate_slider,
                dnsmos_slider,
                speaker_noised_checkbox,
                unconditional_keys,
            ],
        )

        # Generate audio on button click
        generate_button.click(
            fn=generate_audio,
            inputs=[
                model_choice,
                text,
                language,
                speaker_audio,
                prefix_audio,
                emotion1,
                emotion2,
                emotion3,
                emotion4,
                emotion5,
                emotion6,
                emotion7,
                emotion8,
                vq_single_slider,
                fmax_slider,
                pitch_std_slider,
                speaking_rate_slider,
                dnsmos_slider,
                speaker_noised_checkbox,
                cfg_scale_slider,
                min_p_slider,
                seed_number,
                randomize_seed_toggle,
                unconditional_keys,
            ],
            outputs=[output_audio, seed_number],
        )

    return demo


if __name__ == "__main__":
    demo = build_interface()
    share = getenv("GRADIO_SHARE", "False").lower() in ("true", "1", "t")
    demo.launch(server_name="0.0.0.0", server_port=7860, share=share)



# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import contextlib
from typing import Callable, Dict, List, Tuple, Union

import torch
from torchdiffeq import odeint

from .model.dit import DiT
from .model.t2w_cfm import CodecCFM
from .model.utils import exists, load_checkpoint
from .modeling import Qwen2Code2wavBigvgan


class CudaGraphRunner:
    def __init__(self, fn, device):
        """
        initialize CUDA Graph Wrapper.

        Args:
            original_fast_forward (callable): original fast_forward method.
            device (torch.device): CUDA device.
        """
        torch._dynamo.config.cache_size_limit = 64
        self.fn_compile = torch.compile(
            fn,
            mode="default",
            fullgraph=False,
        )
        self.cuda_graph: Dict[Tuple[int, int], torch.cuda.CUDAGraph] = dict()
        self.input_buffers: Dict[Tuple[int, int], Dict[str, torch.Tensor]] = dict()
        self.output_buffers: Dict[Tuple[int, int], torch.Tensor] = dict()

        self.device = device

        # Create customized CUDA stream for Cuda Graph capture
        self.capture_stream = torch.cuda.Stream(device=self.device)

    def capture_cuda_graph(self, x, cond, spk, text, time, mask):
        """
        Capture CUDA Graph。

        Args:
            x (torch.Tensor): nosied input audio.
            cond (torch.Tensor): masked cond audio.
            spk (torch.Tensor): spk embedding.
            text (torch.Tensor): text.
            time (torch.Tensor): time step.
            mask (torch.Tensor or None): mask.
        """
        # Move the input data to the specified device and detach it from the computation graph.

        size = (text.size(0), text.size(1))

        with torch.no_grad():
            if size not in self.input_buffers:
                self.input_buffers[size] = {
                    "x": x.to(self.device, non_blocking=True).clone().detach(),
                    "cond": cond.to(self.device, non_blocking=True).clone().detach(),
                    "spk": spk.to(self.device, non_blocking=True).clone().detach(),
                    "text": text.to(self.device, non_blocking=True).clone().detach(),
                    "time": time.to(self.device, non_blocking=True).clone().detach(),
                    "mask": mask.to(self.device, non_blocking=True).clone().detach()
                    if mask is not None
                    else None,
                }

        # Determine the output shape through a single forward pass and pre-allocate the output buffer.
        with torch.no_grad():
            if size not in self.output_buffers:
                generated = self.fn_compile(
                    x=self.input_buffers[size]["x"],
                    cond=self.input_buffers[size]["cond"],
                    spk=self.input_buffers[size]["spk"],
                    text=self.input_buffers[size]["text"],
                    time=self.input_buffers[size]["time"],
                    mask=self.input_buffers[size]["mask"],
                )
                self.output_buffers[size] = torch.empty_like(generated, device=self.device)

        # Ensure that all previous operations are complete.
        torch.cuda.synchronize(self.device)

        # Begin to capture CUDA Graph
        self.cuda_graph[size] = torch.cuda.CUDAGraph()
        with torch.cuda.graph(self.cuda_graph[size], stream=self.capture_stream):
            # Perform the forward pass and copy the results to the pre-allocated buffer.
            generated = self.fn_compile(
                x=self.input_buffers[size]["x"],
                cond=self.input_buffers[size]["cond"],
                spk=self.input_buffers[size]["spk"],
                text=self.input_buffers[size]["text"],
                time=self.input_buffers[size]["time"],
                mask=self.input_buffers[size]["mask"],
            )
            self.output_buffers[size].copy_(generated)

        # Make sure capture complete
        torch.cuda.synchronize(self.device)

    def __call__(self, x, cond, spk, text, time, mask=None):
        """
        Args:
            x (torch.Tensor): nosied input audio.
            cond (torch.Tensor): masked cond audio.
            spk (torch.Tensor): spk embedding.
            text (torch.Tensor): text.
            time (torch.Tensor): time step.
            mask (torch.Tensor or None): mask.

        Returns:
            torch.Tensor: generated
        """

        size = (text.size(0), text.size(1))

        if size not in self.cuda_graph:
            # Capture the CUDA Graph on the first call.
            self.capture_cuda_graph(x, cond, spk, text, time, mask)

        # Update input to buffer.
        self.input_buffers[size]["x"].copy_(x.to(self.device, non_blocking=True))
        self.input_buffers[size]["cond"].copy_(cond.to(self.device, non_blocking=True))
        self.input_buffers[size]["spk"].copy_(spk.to(self.device, non_blocking=True))
        self.input_buffers[size]["text"].copy_(text.to(self.device, non_blocking=True))
        self.input_buffers[size]["time"].copy_(time.to(self.device, non_blocking=True))
        if self.input_buffers[size]["mask"] is not None and mask is not None:
            self.input_buffers[size]["mask"].copy_(mask.to(self.device, non_blocking=True))
        elif mask is None:
            self.input_buffers[size]["mask"] = None
        # Replay CUDA Graph
        self.cuda_graph[size].replay()
        return self.output_buffers[size]


class BatchCodecCFM(CodecCFM):
    @torch.no_grad()
    def fast_block_sample(
        self,
        cond: float["b n d"] | float["b nw"],  # noqa: F722
        codec: int["b nc dc"],
        ref_mel: float["b n d"],  # noqa: F722
        y0: float["b n d"],
        lens: int[b] | None = None,  # noqa: F821
        steps=32,
        cfg_strength=1.0,
        sway_sampling_coef=None,
        seed: int | None = None,
        max_duration=4096,
        vocoder: Callable[[float["b d n"]], float["b nw"]] | None = None,  # noqa: F722
        no_ref_audio=False,
        duplicate_test=False,
        t_inter=0.1,
        edit_mask=None,
    ):
        self.eval()

        max_duration = y0.shape[1]
        if next(self.parameters()).dtype == torch.float16:
            cond = cond.half()
            ref_mel = ref_mel.half()
            y0 = y0.half()
            # print(next(self.parameters()).dtype)

        # raw wave

        cond = cond.unsqueeze(1).repeat(1, max_duration, 1)
        batch, cond_seq_len, device = *ref_mel.shape[:2], cond.device
        if not exists(lens):
            lens = torch.full((batch,), cond_seq_len, device=device, dtype=torch.long)

        mask = None

        # test for no ref audio
        if no_ref_audio:
            cond = torch.zeros_like(cond)

        # neural ode

        def fn(t, x):
            out_put = self.transformer.fast_forward(
                x=x,
                text=codec,
                spk=cond,
                cond=ref_mel,
                time=t,
                mask=mask,
            )
            pred, null_pred = torch.chunk(out_put, 2, dim=0)
            return pred + (pred - null_pred) * cfg_strength

        t_start = 0
        t = torch.linspace(t_start, 1, steps, device=self.device, dtype=ref_mel.dtype)
        if sway_sampling_coef is not None:
            t = t + sway_sampling_coef * (torch.cos(torch.pi / 2 * t) - 1 + t)

        trajectory = odeint(fn, y0, t, **self.odeint_kwargs)

        sampled = trajectory[-1]
        out = sampled
        # out = torch.where(cond_mask, ref_mel, out)
        return out, trajectory


class Qwen2Code2wavDit(torch.nn.Module):
    def __init__(self, ckpt, frequency: str = "50hz", device="cpu"):
        super().__init__()
        self.frequency = frequency
        self.device = device
        self.dit = DiT(
            dim=1024,
            depth=22 if frequency == "50hz" else 32,
            heads=16,
            ff_mult=2,
            text_dim=512,
            conv_layers=4,
            use_codec=True,
            repeats=2 if frequency == "50hz" else 4,
            attn_processor="stream_block_sr" if frequency == "50hz" else "stream_block_8_L_4",
            text_num_embeds=8193 if frequency == "50hz" else 32769,
            mel_dim=80,
        )
        self.mel_spec_kwargs = dict(
            target_sample_rate=16000,
            n_mel_channels=80,
            hop_length=160,
        )
        self.odeint_kwargs = dict(
            method="euler",
        )
        self.cfm_model = BatchCodecCFM(
            transformer=self.dit,
            mel_spec_kwargs=self.mel_spec_kwargs,
            odeint_kwargs=self.odeint_kwargs,
        ).to(device)
        self.cfm_model = load_checkpoint(self.cfm_model, ckpt, device, use_ema=True)

    def sample(self, cond, ref_mel, codec, steps=10, cfg_strength=0.5, sway_sampling_coef=-1.0):
        y_all = torch.randn([1, 30000, 80], device=self.device, dtype=ref_mel.dtype)
        expect_y_len = codec.shape[1] * (2 if self.frequency == "50hz" else 4)
        y0 = y_all[:, :expect_y_len]
        with torch.inference_mode():
            generated, _ = self.cfm_model.sample(
                cond=cond,
                ref_mel=ref_mel,
                codec=codec,
                steps=steps,
                cfg_strength=cfg_strength,
                sway_sampling_coef=sway_sampling_coef,
                y0=y0,
            )
        generated = generated.to(torch.float32)
        generated_mel_spec = generated.permute(0, 2, 1)
        return generated_mel_spec

    def fast_block_sample(
        self,
        cond,
        codec,
        ref_mel,
        y0,
        steps=10,
        cfg_strength=0.5,
        sway_sampling_coef=-1.0,
    ):
        return self.cfm_model.fast_block_sample(
            cond=cond,
            codec=codec,
            ref_mel=ref_mel,
            y0=y0,
            steps=steps,
            cfg_strength=cfg_strength,
            sway_sampling_coef=sway_sampling_coef,
        )


class Qwen2Code2wav(torch.nn.Module):
    def __init__(
        self,
        dit_ckpt,
        bigvgan_ckpt,
        steps: int = 10,
        bs_mel: int = 24,
        odeint_method: str = "euler",
        odeint_method_relaxed: bool = False,
        batched_chunk: int = 3,
        frequency: str = "50hz",
        device="cpu",
        with_weight_norm: bool = True,
    ):
        super().__init__()
        self.frequency = frequency
        self.code2wav_dit_model = Qwen2Code2wavDit(
            ckpt=dit_ckpt, frequency=frequency, device=device
        )
        self.code2wav_bigvgan_model = Qwen2Code2wavBigvgan(
            ckpt=bigvgan_ckpt, frequency=frequency, device=device, with_weight_norm=with_weight_norm
        )
        self.device = device

        # odeint method: use ruler for first and last chunk to optimize performance
        self.odeint_method_relaxed = odeint_method_relaxed

        # cfm model: override the odeint method
        self.code2wav_dit_model.cfm_model.odeint_kwargs["method"] = odeint_method

        # dit autocast
        self.code2wav_dit_model.dit.fast_forward = torch.autocast(
            device_type="cuda",
            dtype=torch.bfloat16,
        )(self.code2wav_dit_model.dit.fast_forward)

        self.dit_forward = self.code2wav_dit_model.dit.fast_forward
        self.dit_forward_compiled = self.code2wav_dit_model.dit.fast_forward
        self.dit_forward_compiled_first_chunk = self.code2wav_dit_model.dit.fast_forward
        self.dit_forward_cudagraph_first = self.code2wav_dit_model.dit.fast_forward
        self.dit_forward_cudagraph_intermediate = self.code2wav_dit_model.dit.fast_forward
        self.dit_forward_cudagraph_last = self.code2wav_dit_model.dit.fast_forward

        self.torch_compile_first_chunk = False

        self.factor = 2 if frequency == "50hz" else 4
        self.steps = steps
        self.bs_mel = bs_mel
        self.bs_codec = bs_mel // self.factor
        self.past_cache_size = bs_mel * self.factor
        self.future_cache_size = bs_mel * 1
        self.chunk_size = bs_mel * batched_chunk
        self.future_size = 20 if self.frequency == "50hz" else 13
        self.past_size = 20 if self.frequency == "50hz" else 51

        text_embed = self.code2wav_dit_model.dit.text_embed
        if hasattr(text_embed, "codec_embed"):
            self.codec_embed_size = text_embed.codec_embed.weight.size(0)
        elif hasattr(text_embed, "text_embed"):
            self.codec_embed_size = text_embed.text_embed.weight.size(0)
        else:
            self.codec_embed_size = -1

    @contextlib.contextmanager
    def relax_odeint_method(self, relax: bool = False):
        if relax and self.odeint_method_relaxed:
            odeint_method = self.code2wav_dit_model.cfm_model.odeint_kwargs["method"]
            self.code2wav_dit_model.cfm_model.odeint_kwargs["method"] = "euler"
        yield
        if relax and self.odeint_method_relaxed:
            self.code2wav_dit_model.cfm_model.odeint_kwargs["method"] = odeint_method

    def enable_torch_compile(self, compile_first_chunk: bool = False):
        self.torch_compile_first_chunk = compile_first_chunk

        self.dit_forward_compiled = torch.compile(
            self.code2wav_dit_model.dit.fast_forward,
            # mode="default",
            mode="reduce-overhead",
            fullgraph=False,
        )
        self.dit_forward_cudagraph_first = CudaGraphRunner(
            self.code2wav_dit_model.dit.fast_forward, self.device
        )
        self.dit_forward_cudagraph_intermediate = CudaGraphRunner(
            self.code2wav_dit_model.dit.fast_forward, self.device
        )
        self.dit_forward_cudagraph_last = CudaGraphRunner(
            self.code2wav_dit_model.dit.fast_forward, self.device
        )

    @torch.inference_mode()
    def forward(
        self,
        cond,
        ref_mel,
        codec,
        steps=10,
        cfg_strength=0.5,
        sway_sampling_coef=-1.0,
    ):
        generated_mel = self.code2wav_dit_model.sample(
            cond,
            ref_mel,
            codec,
            steps=steps,
            cfg_strength=cfg_strength,
            sway_sampling_coef=sway_sampling_coef,
        )
        generated_mel = generated_mel.permute(0, 2, 1)
        waveform = self.code2wav_bigvgan_model(generated_mel)
        return waveform

    @torch.inference_mode()
    def process_chunk_dit_batch(
        self,
        cond,
        ref_mel,
        codec,
        y0,
        steps,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        if self.codec_embed_size > 0:
            codec[codec >= self.codec_embed_size] = 0

        self.code2wav_dit_model.dit.fast_forward = self.dit_forward_cudagraph_intermediate
        generated, _ = self.code2wav_dit_model.fast_block_sample(
            cond=cond,
            codec=codec,
            ref_mel=ref_mel,
            y0=y0,
            steps=steps,
            cfg_strength=0.5,
            sway_sampling_coef=-1.0,
        )
        return generated

    @torch.inference_mode()
    def process_chunk_bigvgan_batch(self, mel_batch):
        return self.code2wav_bigvgan_model(mel_batch)

    @torch.inference_mode()
    def process_little_chunk(
        self,
        cond,
        ref_mel,
        codec_all,
        y_all,
        i,
        steps,
        prev_generated: torch.Tensor,
        finished: bool = False,
        cfg_strength=0.5,
        sway_sampling_coef=-1.0,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        # mask to prevent codec from being out of range (the eos token)
        if self.codec_embed_size > 0:
            codec_all[codec_all >= self.codec_embed_size] = 0

        return None, self.forward(
            cond,
            ref_mel,
            codec_all,
            steps=steps,
            cfg_strength=cfg_strength,
            sway_sampling_coef=sway_sampling_coef,
        )

    @torch.inference_mode()
    def process_chunk(
        self,
        cond,
        ref_mel,
        codec_all,
        y_all,
        i,
        steps,
        prev_generated: Union[torch.Tensor, List[torch.Tensor]],
        finished: bool = False,
        cfg_strength=0.5,
        sway_sampling_coef=-1.0,
    ) -> Tuple[Union[torch.Tensor, List[torch.Tensor]], torch.Tensor]:
        start_index = max(i * self.chunk_size - self.past_cache_size, 0)
        end_index = min(
            (i + 1) * self.chunk_size + self.future_cache_size, codec_all.shape[1] * self.factor
        )
        y0 = y_all[:, start_index:end_index].reshape(1, -1, 80).contiguous()
        codec = (
            codec_all[:, start_index // self.factor : end_index // self.factor]
            .reshape(1, -1)
            .contiguous()
        )

        # mask to prevent codec from being out of range (the eos token)
        if self.codec_embed_size > 0:
            codec[codec >= self.codec_embed_size] = 0

        # N.B. when using cuda graph ("reduce-overhead" mode), don't compile
        # shape for the first and the last chunk, as it will affect the performance
        # for normal chunks.
        #
        # The reason is not clear yet. The default torch.compile() mode is not affected.
        if i == 0:
            if self.torch_compile_first_chunk:
                self.code2wav_dit_model.dit.fast_forward = self.dit_forward_compiled
            else:
                self.code2wav_dit_model.dit.fast_forward = self.dit_forward_cudagraph_first
        elif finished:
            self.code2wav_dit_model.dit.fast_forward = self.dit_forward_cudagraph_last
        else:
            self.code2wav_dit_model.dit.fast_forward = self.dit_forward_cudagraph_intermediate

        with self.relax_odeint_method(relax=i == 0 or finished):
            generated, _ = self.code2wav_dit_model.fast_block_sample(
                cond=cond,
                codec=codec,
                ref_mel=ref_mel,
                y0=y0,
                steps=steps,
                cfg_strength=cfg_strength,
                sway_sampling_coef=sway_sampling_coef,
            )

        if self.frequency == "50hz":
            return self.process_chunk_for_50hz(
                i,
                start_index,
                end_index,
                finished,
                prev_generated,
                generated,
            )
        else:
            raise ValueError(f"Unsupported frequency: {self.frequency}")

    def process_chunk_for_50hz(
        self,
        i: int,
        start_index: int,
        end_index: int,
        finished: bool,
        prev_generated: torch.Tensor,
        generated: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        if i == 0:
            generated = generated.to(torch.float32)[:, : self.chunk_size, :]
            mel = generated
        elif finished:
            generated = generated.to(torch.float32)[:, self.past_cache_size :, :]
            mel = torch.cat([prev_generated[:, -self.future_size * 2 :, :], generated], dim=1)
        else:
            # Note that self.chunk_size == self.past_cache_size, so the following branch
            # can be simplified. But for clearness, we keep it as it is in transformers.
            if start_index == 0:
                generated = generated.to(torch.float32)[
                    :, i * self.chunk_size : -self.future_cache_size, :
                ]
            else:
                generated = generated.to(torch.float32)[
                    :, self.past_cache_size : -self.future_cache_size, :
                ]
            mel = torch.cat([prev_generated[:, -self.future_size * 2 :, :], generated], dim=1)

        audio = self.code2wav_bigvgan_model(mel)
        if i == 0:
            audio_output = audio[: -self.future_size * 240]
        elif finished:
            audio_output = audio[self.future_size * 240 :]
        else:
            audio_output = audio[self.future_size * 240 : -self.future_size * 240]
        return generated, audio_output



import whisper
from whisper_timestamped import transcribe_timestamped


_MODELS = {
    "tiny.en": "https://openaipublic.azureedge.net/main/whisper/models/d3dd57d32accea0b295c96e26691aa14d8822fac7d9d27d5dc00b4ca2826dd03/tiny.en.pt",
    "tiny": "https://openaipublic.azureedge.net/main/whisper/models/65147644a518d12f04e32d6f3b26facc3f8dd46e5390956a9424a650c0ce22b9/tiny.pt",
    "base.en": "https://openaipublic.azureedge.net/main/whisper/models/25a8566e1d0c1e2231d1c762132cd20e0f96a85d16145c3a00adf5d1ac670ead/base.en.pt",
    "base": "https://openaipublic.azureedge.net/main/whisper/models/ed3a0b6b1c0edf879ad9b11b1af5a0e6ab5db9205f891f668f8b0e6c6326e34e/base.pt",
    "small.en": "https://openaipublic.azureedge.net/main/whisper/models/f953ad0fd29cacd07d5a9eda5624af0f6bcf2258be67c92b79389873d91e0872/small.en.pt",
    "small": "https://openaipublic.azureedge.net/main/whisper/models/9ecf779972d90ba49c06d968637d720dd632c55bbf19d441fb42bf17a411e794/small.pt",
    "medium.en": "https://openaipublic.azureedge.net/main/whisper/models/d7440d1dc186f76616474e0ff0b3b6b879abc9d1a4926b7adfa41db2d497ab4f/medium.en.pt",
    "medium": "https://openaipublic.azureedge.net/main/whisper/models/345ae4da62f9b3d59415adc60127b97c714f32e89e936602e85993674d08dcb1/medium.pt",
    "large-v1": "https://openaipublic.azureedge.net/main/whisper/models/e4b87e7e0bf463eb8e6956e646f1e277e901512310def2c24bf0e11bd3c28e9a/large-v1.pt",
    "large-v2": "https://openaipublic.azureedge.net/main/whisper/models/81f7c96c852ee8fc832187b0132e569d6c3065a3252ed18e56effd0b6a73e524/large-v2.pt",
    "large-v3": "https://openaipublic.azureedge.net/main/whisper/models/e5b1a55b89c1367dacf97e3e19bfd829a01529dbfdeefa8caeb59b3f1b81dadb/large-v3.pt",
    "large": "https://openaipublic.azureedge.net/main/whisper/models/e5b1a55b89c1367dacf97e3e19bfd829a01529dbfdeefa8caeb59b3f1b81dadb/large-v3.pt",
}

# whisper records/tmp.wav --language zh --model tiny --model_dir ./models
# whisper records/tmp.wav --language zh --model base --model_dir ./models
# whisper records/tmp.wav --language zh --model small --model_dir ./models
# whisper records/tmp.wav --language zh --model medium --model_dir ./models
# whisper records/tmp.wav --language zh --model large --model_dir ./models


def whisper_transcribe(audio_path, download_root, model_size="base", target_lang="zh"):
    audio_model = whisper.load_model(model_size, download_root=download_root)
    result = audio_model.transcribe(
        audio_path,
        language=target_lang,
        beam_size=5,
        word_timestamps=True,
        condition_on_previous_text=True,
    )
    print(result)
    text = result["text"].strip()
    print(text)


# whisper_timestamped records/tmp.wav --language zh --model base --model_dir ./models
# whisper_timestamped records/tmp.wav --language zh --model base --model_dir ./models
# whisper_timestamped records/tmp.wav --language zh --model small --model_dir ./models
# whisper_timestamped records/tmp.wav --language zh --model medium --model_dir ./models
# whisper_timestamped records/tmp.wav --language zh --model large --model_dir ./models


def whisper_transcribe_timestamped(audio_path, download_root, model_size="base", target_lang="zh"):
    audio_model = whisper.load_model(model_size, download_root=download_root)
    # help(transcribe_timestamped)
    result = transcribe_timestamped(audio_model, audio_path, language=target_lang)
    print(result)
    text = result["text"].strip()
    print(text)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model_type",
        "-t",
        type=str,
        default="whisper",
        help="choice whisper | whisper_timestamped",
    )
    parser.add_argument(
        "--audio_path", "-a", type=str, default="./records/tmp.wav", help="audio path"
    )
    parser.add_argument("--model_size", "-s", type=str, default="base", help="model size")
    parser.add_argument("--model_path", "-m", type=str, default="./models", help="model root path")
    parser.add_argument("--lang", "-l", type=str, default="zh", help="target language")
    args = parser.parse_args()
    if args.model_type == "whisper_timestamped":
        whisper_transcribe_timestamped(args.audio_path, args.model_path, args.model_size, args.lang)
    else:
        whisper_transcribe(args.audio_path, args.model_path, args.model_size, args.lang)



v1 = {
    "resblock": "1",
    "num_gpus": 0,
    "batch_size": 16,
    "learning_rate": 0.0004,
    "adam_b1": 0.8,
    "adam_b2": 0.99,
    "lr_decay": 0.999,
    "seed": 1234,
    "upsample_rates": [8, 8, 2, 2],
    "upsample_kernel_sizes": [16, 16, 4, 4],
    "upsample_initial_channel": 512,
    "resblock_kernel_sizes": [3, 7, 11],
    "resblock_dilation_sizes": [[1, 3, 5], [1, 3, 5], [1, 3, 5]],
    "resblock_initial_channel": 256,
    "segment_size": 8192,
    "num_mels": 80,
    "num_freq": 1025,
    "n_fft": 1024,
    "hop_size": 256,
    "win_size": 1024,
    "sampling_rate": 22050,
    "fmin": 0,
    "fmax": 8000,
    "fmax_loss": None,
    "num_workers": 4,
    "dist_config": {"dist_backend": "nccl", "dist_url": "tcp://localhost:54321", "world_size": 1},
}



# Copyright (c) 2024 Alibaba Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Iterable
from typing import List


class AbsTokenizer(ABC):
    @abstractmethod
    def text2tokens(self, line: str) -> List[str]:
        raise NotImplementedError

    @abstractmethod
    def tokens2text(self, tokens: Iterable[str]) -> str:
        raise NotImplementedError

    
    
    def encode(self, line: str, **kwargs) -> List[str]:

        return self.text2tokens(line)


import os
from livekit import api
import asyncio


async def create_room():
    # will automatically use the LIVEKIT_URL, LIVEKIT_API_KEY and LIVEKIT_API_SECRET env vars
    lkapi = api.LiveKitAPI()
    room_info = await lkapi.room.create_room(
        api.CreateRoomRequest(name=os.getenv("ROOM_NAME", "chat-room")),
    )
    print("room_info", room_info)
    results = await lkapi.room.list_rooms(api.ListRoomsRequest(names=["room", "chat-room"]))
    print("room_list", results.rooms, type(results))
    await lkapi.aclose()


if __name__ == "__main__":
    asyncio.get_event_loop().run_until_complete(create_room())



import os
import logging

import wave
import unittest
from apipeline.pipeline.pipeline import Pipeline, FrameDirection
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.pipeline.runner import PipelineRunner
from apipeline.processors.logger import FrameLogger
from apipeline.frames import TextFrame, AudioRawFrame, EndFrame

from src.processors.speech.audio_save_processor import AudioSaveProcessor
from src.processors.voice.glm_voice_processor import GLMAudioVoiceProcessor, GLMTextVoiceProcessor
from src.common.types import TEST_DIR
from src.common.logger import Logger
from src.types.frames import PathAudioRawFrame

from dotenv import load_dotenv

load_dotenv(override=True)

"""
LOG_LEVEL=debug python -m unittest test.integration.processors.test_glm_voice_processor.TestProcessor.test_audio
LOG_LEVEL=debug GLM_VOICE_TYPE=text python -m unittest test.integration.processors.test_glm_voice_processor.TestProcessor.test_text
"""


class TestProcessor(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False)
        cls.glm_voice_type = os.getenv("GLM_VOICE_TYPE", "audio")
        cls.test_file = os.path.join(TEST_DIR, "audio_files/hi.wav")
        cls.args = {
            "bnb_quant_type": "int4",
            "device": "cuda",
            "lm_gen_args": {"max_new_token": 2000, "temperature": 0.2, "top_p": 0.8},
            "model_path": "./models/THUDM/glm-4-voice-9b",
            "torch_dtype": "auto",
            "voice_decoder_path": "./models/THUDM/glm-4-voice-decoder",
            "voice_out_args": {"audio_channels": 1, "audio_sample_rate": 22050},
            "voice_tokenizer_path": "./models/THUDM/glm-4-voice-tokenizer",
        }

    @classmethod
    def tearDownClass(cls):
        pass

    async def asyncSetUp(self):
        if self.glm_voice_type == "text":
            voice_processor = GLMTextVoiceProcessor(**self.args)
        else:
            voice_processor = GLMAudioVoiceProcessor(**self.args)

        pipeline = Pipeline(
            [
                FrameLogger(include_frame_types=[AudioRawFrame, TextFrame]),
                voice_processor,  # output TextFrame and AudioRawFrame
                FrameLogger(include_frame_types=[TextFrame, AudioRawFrame]),
                AudioSaveProcessor(prefix_name="bot_speak"),
                # FrameLogger(include_frame_types=[TextFrame, AudioRawFrame]),
            ]
        )

        self.task = PipelineTask(
            pipeline,
            PipelineParams(
                allow_interruptions=False,  # close pipeline interruptions, use model interrupt
                enable_metrics=True,
            ),
        )

    async def asyncTearDown(self):
        pass

    async def test_audio(self):
        runner = PipelineRunner()

        with wave.open(self.test_file, "rb") as wav_file:
            audio_frame = AudioRawFrame(
                audio=wav_file.readframes(wav_file.getnframes()),
                sample_rate=wav_file.getframerate(),
                num_channels=wav_file.getnchannels(),
                sample_width=wav_file.getsampwidth(),
            )
            path_frame = PathAudioRawFrame(
                path=self.test_file,
                audio=wav_file.readframes(wav_file.getnframes()),
                sample_rate=wav_file.getframerate(),
                num_channels=wav_file.getnchannels(),
                sample_width=wav_file.getsampwidth(),
            )

        await self.task.queue_frames(
            [
                audio_frame,
                path_frame,
                EndFrame(),
            ]
        )
        await runner.run(self.task)

    async def test_text(self):
        runner = PipelineRunner()

        await self.task.queue_frames(
            [
                TextFrame("你好"),
                TextFrame("你叫什么名字"),
                EndFrame(),
            ]
        )
        await runner.run(self.task)



# Copyright (c) 2024 Alibaba Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer
from inspiremusic.utils.mask import make_pad_mask
from inspiremusic.utils.hinter import hint_once
from inspiremusic.utils.common import DTYPES

class QwenEncoder(nn.Module):
    def __init__(
            self,
            input_size: int,
            dtype: str = "fp16",
            pretrain_path: str = "Qwen/Qwen2.0-0.5B",
            trainable: bool = False,
            do_fusion_emb: bool = False,
            fusion_drop_rate: float = 0.0,
    ):
        super(QwenEncoder, self).__init__()
        self.input_size = input_size
        self.trainable = trainable
        self.dtype = DTYPES.get(dtype, torch.float32)
        self.model = AutoModelForCausalLM.from_pretrained(pretrain_path, device_map="auto", attn_implementation="flash_attention_2", torch_dtype=self.dtype)
        self._output_size = self.model.config.hidden_size
        self.do_fusion_emb = do_fusion_emb
        self.hidden_norm = torch.nn.LayerNorm(self._output_size)
        self.fusion_dropout = nn.Dropout(fusion_drop_rate)
        if do_fusion_emb:
            self.fusion_layer = torch.nn.Linear(self._output_size * 2, self._output_size)
            self.emb_norm = torch.nn.LayerNorm(self._output_size)
            self.fusion_norm = torch.nn.LayerNorm(self._output_size)
            from inspiremusic.transformer.activation import Swish
            self.fusion_act = Swish(self)

        if not self.trainable:
            self.model.eval()

    def output_size(self) -> int:
        return self._output_size

    def forward(
            self,
            input_ids: torch.Tensor,
            ilens: torch.Tensor,
    ):
        device = input_ids.device
        input_ids = torch.clamp(input_ids, min=0, max=None)
        input_masks = (~make_pad_mask(ilens)).to(device).long()
        if not self.trainable:
            with torch.no_grad():
                model_outputs = self.model(
                    input_ids=input_ids,
                    attention_mask=input_masks,
                    output_hidden_states=True
                )
        else:
            model_outputs = self.model(
                input_ids=input_ids,
                attention_mask=input_masks,
                output_hidden_states=True
            )
        outs = model_outputs.hidden_states[-1]
        outs = self.hidden_norm(outs)
        if self.do_fusion_emb:
            hint_once("fuse embedding and LM outputs", "fuse_emb")
            outs = self.fusion_dropout(self.fusion_act(outs))
            emb = model_outputs.hidden_states[0]
            emb = self.fusion_dropout(self.fusion_act(self.emb_norm(emb)))
            outs = self.fusion_layer(
                torch.cat([outs, emb], dim=-1)
            )
            outs = self.fusion_act(self.fusion_norm(outs))

        return outs, ilens


class QwenEmbeddingEncoder(nn.Module):
    def __init__(
            self,
            input_size: int,
            dtype: str = "fp16",
            pretrain_path: str = "Qwen/Qwen2.0-0.5B",
    ):
        super(QwenEmbeddingEncoder, self).__init__()
        self.input_size = input_size
        self.dtype = DTYPES.get(dtype, torch.float32)
        from transformers import Qwen2ForCausalLM
        self.model = Qwen2ForCausalLM.from_pretrained(pretrain_path, device_map="auto", attn_implementation="flash_attention_2", torch_dtype=self.dtype)
        self._output_size = self.model.config.hidden_size

    def output_size(self) -> int:
        return self._output_size

    def forward(
            self,
            input_embeds: torch.Tensor,
            ilens: torch.Tensor,
    ):
        input_masks = (~make_pad_mask(ilens)).to(input_embeds.device).long()

        outs = self.model(
            inputs_embeds=input_embeds,
            attention_mask=input_masks,
            output_hidden_states=True,
            return_dict=True,
        )

        return outs.hidden_states[-1], input_masks

    def forward_one_step(self, xs, masks, cache=None):

        outs = self.model(
            inputs_embeds=xs,
            attention_mask=masks,
            output_hidden_states=True,
            return_dict=True,
            use_cache=True,
            past_key_values=cache,
        )
        xs = outs.hidden_states[-1]
        new_cache = outs.past_key_values

        return xs, masks, new_cache


class QwenInputOnlyEncoder(nn.Module):
    def __init__(
            self,
            input_size: int,
            dtype: str = "fp16",
            pretrain_path: str = "Qwen/Qwen2.0-0.5B",
    ):
        super(QwenInputOnlyEncoder, self).__init__()
        self.input_size = input_size
        self.dtype = DTYPES.get(dtype, torch.float32)
        from transformers import Qwen2ForCausalLM
        model = Qwen2ForCausalLM.from_pretrained(pretrain_path, device_map="auto", attn_implementation="flash_attention_2", torch_dtype=self.dtype)
        self.embed = model.model.embed_tokens
        for p in self.embed.parameters():
            p.requires_grad = False
            # set text embedding to non-trainable

        # self.post_embed = model.model.rotary_emb
        self._output_size = model.config.hidden_size

    def output_size(self) -> int:
        return self._output_size

    def forward(
            self,
            input_ids: torch.Tensor,
            ilens: torch.Tensor,
    ):
        input_masks = (~make_pad_mask(ilens)).to(input_ids.device).long()

        outs = self.embed(input_ids)

        return outs, input_masks
                                                       


#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
# Copyright FunASR (https://github.com/alibaba-damo-academy/FunASR). All Rights Reserved.
#  MIT License  (https://opensource.org/licenses/MIT)

import time
import torch
import torch.nn as nn
import torch.functional as F
import logging
from typing import Dict, Tuple
from contextlib import contextmanager
from distutils.version import LooseVersion

from funasr_detach.register import tables
from funasr_detach.models.ctc.ctc import CTC
from funasr_detach.utils import postprocess_utils
from funasr_detach.metrics.compute_acc import th_accuracy
from funasr_detach.utils.datadir_writer import DatadirWriter
from funasr_detach.models.paraformer.model import Paraformer
from funasr_detach.models.paraformer.search import Hypothesis
from funasr_detach.models.paraformer.cif_predictor import mae_loss
from funasr_detach.train_utils.device_funcs import force_gatherable
from funasr_detach.losses.label_smoothing_loss import LabelSmoothingLoss
from funasr_detach.models.transformer.utils.add_sos_eos import add_sos_eos
from funasr_detach.models.transformer.utils.nets_utils import make_pad_mask, pad_list
from funasr_detach.utils.load_utils import load_audio_text_image_video, extract_fbank
from funasr_detach.models.scama.utils import sequence_mask

if LooseVersion(torch.__version__) >= LooseVersion("1.6.0"):
    from torch.cuda.amp import autocast
else:
    # Nothing to do if torch<1.6.0
    @contextmanager
    def autocast(enabled=True):
        yield


@tables.register("model_classes", "SCAMA")
class SCAMA(nn.Module):
    """
    Author: Shiliang Zhang, Zhifu Gao, Haoneng Luo, Ming Lei, Jie Gao, Zhijie Yan, Lei Xie
    SCAMA: Streaming chunk-aware multihead attention for online end-to-end speech recognition
    https://arxiv.org/abs/2006.01712
    """

    def __init__(
        self,
        specaug: str = None,
        specaug_conf: dict = None,
        normalize: str = None,
        normalize_conf: dict = None,
        encoder: str = None,
        encoder_conf: dict = None,
        decoder: str = None,
        decoder_conf: dict = None,
        ctc: str = None,
        ctc_conf: dict = None,
        ctc_weight: float = 0.5,
        predictor: str = None,
        predictor_conf: dict = None,
        predictor_bias: int = 0,
        predictor_weight: float = 0.0,
        input_size: int = 80,
        vocab_size: int = -1,
        ignore_id: int = -1,
        blank_id: int = 0,
        sos: int = 1,
        eos: int = 2,
        lsm_weight: float = 0.0,
        length_normalized_loss: bool = False,
        share_embedding: bool = False,
        **kwargs,
    ):

        super().__init__()

        if specaug is not None:
            specaug_class = tables.specaug_classes.get(specaug)
            specaug = specaug_class(**specaug_conf)

        if normalize is not None:
            normalize_class = tables.normalize_classes.get(normalize)
            normalize = normalize_class(**normalize_conf)

        encoder_class = tables.encoder_classes.get(encoder)
        encoder = encoder_class(input_size=input_size, **encoder_conf)
        encoder_output_size = encoder.output_size()

        decoder_class = tables.decoder_classes.get(decoder)
        decoder = decoder_class(
            vocab_size=vocab_size,
            encoder_output_size=encoder_output_size,
            **decoder_conf,
        )
        if ctc_weight > 0.0:

            if ctc_conf is None:
                ctc_conf = {}

            ctc = CTC(
                odim=vocab_size, encoder_output_size=encoder_output_size, **ctc_conf
            )

        predictor_class = tables.predictor_classes.get(predictor)
        predictor = predictor_class(**predictor_conf)

        # note that eos is the same as sos (equivalent ID)
        self.blank_id = blank_id
        self.sos = sos if sos is not None else vocab_size - 1
        self.eos = eos if eos is not None else vocab_size - 1
        self.vocab_size = vocab_size
        self.ignore_id = ignore_id
        self.ctc_weight = ctc_weight

        self.specaug = specaug
        self.normalize = normalize

        self.encoder = encoder

        if ctc_weight == 1.0:
            self.decoder = None
        else:
            self.decoder = decoder

        self.criterion_att = LabelSmoothingLoss(
            size=vocab_size,
            padding_idx=ignore_id,
            smoothing=lsm_weight,
            normalize_length=length_normalized_loss,
        )

        if ctc_weight == 0.0:
            self.ctc = None
        else:
            self.ctc = ctc

        self.predictor = predictor
        self.predictor_weight = predictor_weight
        self.predictor_bias = predictor_bias

        self.criterion_pre = mae_loss(normalize_length=length_normalized_loss)

        self.share_embedding = share_embedding
        if self.share_embedding:
            self.decoder.embed = None

        self.length_normalized_loss = length_normalized_loss
        self.beam_search = None
        self.error_calculator = None

        if self.encoder.overlap_chunk_cls is not None:
            from funasr_detach.models.scama.chunk_utilis import (
                build_scama_mask_for_cross_attention_decoder,
            )

            self.build_scama_mask_for_cross_attention_decoder_fn = (
                build_scama_mask_for_cross_attention_decoder
            )
            self.decoder_attention_chunk_type = kwargs.get(
                "decoder_attention_chunk_type", "chunk"
            )

    def forward(
        self,
        speech: torch.Tensor,
        speech_lengths: torch.Tensor,
        text: torch.Tensor,
        text_lengths: torch.Tensor,
        **kwargs,
    ) -> Tuple[torch.Tensor, Dict[str, torch.Tensor], torch.Tensor]:
        """Encoder + Decoder + Calc loss
        Args:
                speech: (Batch, Length, ...)
                speech_lengths: (Batch, )
                text: (Batch, Length)
                text_lengths: (Batch,)
        """

        decoding_ind = kwargs.get("decoding_ind")
        if len(text_lengths.size()) > 1:
            text_lengths = text_lengths[:, 0]
        if len(speech_lengths.size()) > 1:
            speech_lengths = speech_lengths[:, 0]

        batch_size = speech.shape[0]

        # Encoder
        ind = self.encoder.overlap_chunk_cls.random_choice(self.training, decoding_ind)
        encoder_out, encoder_out_lens = self.encode(speech, speech_lengths, ind=ind)

        loss_ctc, cer_ctc = None, None
        loss_pre = None
        stats = dict()

        # decoder: CTC branch

        if self.ctc_weight > 0.0:

            encoder_out_ctc, encoder_out_lens_ctc = (
                self.encoder.overlap_chunk_cls.remove_chunk(
                    encoder_out, encoder_out_lens, chunk_outs=None
                )
            )

            loss_ctc, cer_ctc = self._calc_ctc_loss(
                encoder_out_ctc, encoder_out_lens_ctc, text, text_lengths
            )
            # Collect CTC branch stats
            stats["loss_ctc"] = loss_ctc.detach() if loss_ctc is not None else None
            stats["cer_ctc"] = cer_ctc

        # decoder: Attention decoder branch
        loss_att, acc_att, cer_att, wer_att, loss_pre = self._calc_att_predictor_loss(
            encoder_out, encoder_out_lens, text, text_lengths
        )

        # 3. CTC-Att loss definition
        if self.ctc_weight == 0.0:
            loss = loss_att + loss_pre * self.predictor_weight
        else:
            loss = (
                self.ctc_weight * loss_ctc
                + (1 - self.ctc_weight) * loss_att
                + loss_pre * self.predictor_weight
            )

        # Collect Attn branch stats
        stats["loss_att"] = loss_att.detach() if loss_att is not None else None
        stats["acc"] = acc_att
        stats["cer"] = cer_att
        stats["wer"] = wer_att
        stats["loss_pre"] = loss_pre.detach().cpu() if loss_pre is not None else None

        stats["loss"] = torch.clone(loss.detach())

        # force_gatherable: to-device and to-tensor if scalar for DataParallel
        if self.length_normalized_loss:
            batch_size = (text_lengths + self.predictor_bias).sum()
        loss, stats, weight = force_gatherable((loss, stats, batch_size), loss.device)
        return loss, stats, weight

    def encode(
        self,
        speech: torch.Tensor,
        speech_lengths: torch.Tensor,
        **kwargs,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Encoder. Note that this method is used by asr_inference.py
        Args:
                speech: (Batch, Length, ...)
                speech_lengths: (Batch, )
                ind: int
        """
        with autocast(False):

            # Data augmentation
            if self.specaug is not None and self.training:
                speech, speech_lengths = self.specaug(speech, speech_lengths)

            # Normalization for feature: e.g. Global-CMVN, Utterance-CMVN
            if self.normalize is not None:
                speech, speech_lengths = self.normalize(speech, speech_lengths)

        # Forward encoder
        encoder_out, encoder_out_lens, _ = self.encoder(speech, speech_lengths)
        if isinstance(encoder_out, tuple):
            encoder_out = encoder_out[0]

        return encoder_out, encoder_out_lens

    def encode_chunk(
        self,
        speech: torch.Tensor,
        speech_lengths: torch.Tensor,
        cache: dict = None,
        **kwargs,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Frontend + Encoder. Note that this method is used by asr_inference.py
        Args:
                speech: (Batch, Length, ...)
                speech_lengths: (Batch, )
                ind: int
        """
        with autocast(False):

            # Data augmentation
            if self.specaug is not None and self.training:
                speech, speech_lengths = self.specaug(speech, speech_lengths)

            # Normalization for feature: e.g. Global-CMVN, Utterance-CMVN
            if self.normalize is not None:
                speech, speech_lengths = self.normalize(speech, speech_lengths)

        # Forward encoder
        encoder_out, encoder_out_lens, _ = self.encoder.forward_chunk(
            speech, speech_lengths, cache=cache["encoder"]
        )
        if isinstance(encoder_out, tuple):
            encoder_out = encoder_out[0]

        return encoder_out, torch.tensor([encoder_out.size(1)])

    def calc_predictor_chunk(self, encoder_out, encoder_out_lens, cache=None, **kwargs):
        is_final = kwargs.get("is_final", False)

        return self.predictor.forward_chunk(
            encoder_out, cache["encoder"], is_final=is_final
        )

    def _calc_att_predictor_loss(
        self,
        encoder_out: torch.Tensor,
        encoder_out_lens: torch.Tensor,
        ys_pad: torch.Tensor,
        ys_pad_lens: torch.Tensor,
    ):
        ys_in_pad, ys_out_pad = add_sos_eos(ys_pad, self.sos, self.eos, self.ignore_id)
        ys_in_lens = ys_pad_lens + 1

        encoder_out_mask = sequence_mask(
            encoder_out_lens,
            maxlen=encoder_out.size(1),
            dtype=encoder_out.dtype,
            device=encoder_out.device,
        )[:, None, :]
        mask_chunk_predictor = None
        if self.encoder.overlap_chunk_cls is not None:
            mask_chunk_predictor = (
                self.encoder.overlap_chunk_cls.get_mask_chunk_predictor(
                    None, device=encoder_out.device, batch_size=encoder_out.size(0)
                )
            )
            mask_shfit_chunk = self.encoder.overlap_chunk_cls.get_mask_shfit_chunk(
                None, device=encoder_out.device, batch_size=encoder_out.size(0)
            )
            encoder_out = encoder_out * mask_shfit_chunk
        pre_acoustic_embeds, pre_token_length, pre_alphas, _ = self.predictor(
            encoder_out,
            ys_out_pad,
            encoder_out_mask,
            ignore_id=self.ignore_id,
            mask_chunk_predictor=mask_chunk_predictor,
            target_label_length=ys_in_lens,
        )
        predictor_alignments, predictor_alignments_len = (
            self.predictor.gen_frame_alignments(pre_alphas, encoder_out_lens)
        )

        encoder_chunk_size = self.encoder.overlap_chunk_cls.chunk_size_pad_shift_cur
        attention_chunk_center_bias = 0
        attention_chunk_size = encoder_chunk_size
        decoder_att_look_back_factor = (
            self.encoder.overlap_chunk_cls.decoder_att_look_back_factor_cur
        )
        mask_shift_att_chunk_decoder = (
            self.encoder.overlap_chunk_cls.get_mask_shift_att_chunk_decoder(
                None, device=encoder_out.device, batch_size=encoder_out.size(0)
            )
        )
        scama_mask = self.build_scama_mask_for_cross_attention_decoder_fn(
            predictor_alignments=predictor_alignments,
            encoder_sequence_length=encoder_out_lens,
            chunk_size=1,
            encoder_chunk_size=encoder_chunk_size,
            attention_chunk_center_bias=attention_chunk_center_bias,
            attention_chunk_size=attention_chunk_size,
            attention_chunk_type=self.decoder_attention_chunk_type,
            step=None,
            predictor_mask_chunk_hopping=mask_chunk_predictor,
            decoder_att_look_back_factor=decoder_att_look_back_factor,
            mask_shift_att_chunk_decoder=mask_shift_att_chunk_decoder,
            target_length=ys_in_lens,
            is_training=self.training,
        )

        # try:
        # 1. Forward decoder
        decoder_out, _ = self.decoder(
            encoder_out,
            encoder_out_lens,
            ys_in_pad,
            ys_in_lens,
            chunk_mask=scama_mask,
            pre_acoustic_embeds=pre_acoustic_embeds,
        )

        # 2. Compute attention loss
        loss_att = self.criterion_att(decoder_out, ys_out_pad)
        acc_att = th_accuracy(
            decoder_out.view(-1, self.vocab_size),
            ys_out_pad,
            ignore_label=self.ignore_id,
        )
        # predictor loss
        loss_pre = self.criterion_pre(
            ys_in_lens.type_as(pre_token_length), pre_token_length
        )
        # Compute cer/wer using attention-decoder
        if self.training or self.error_calculator is None:
            cer_att, wer_att = None, None
        else:
            ys_hat = decoder_out.argmax(dim=-1)
            cer_att, wer_att = self.error_calculator(ys_hat.cpu(), ys_pad.cpu())

        return loss_att, acc_att, cer_att, wer_att, loss_pre

    def calc_predictor_mask(
        self,
        encoder_out: torch.Tensor,
        encoder_out_lens: torch.Tensor,
        ys_pad: torch.Tensor = None,
        ys_pad_lens: torch.Tensor = None,
    ):
        # ys_in_pad, ys_out_pad = add_sos_eos(ys_pad, self.sos, self.eos, self.ignore_id)
        # ys_in_lens = ys_pad_lens + 1
        ys_out_pad, ys_in_lens = None, None

        encoder_out_mask = sequence_mask(
            encoder_out_lens,
            maxlen=encoder_out.size(1),
            dtype=encoder_out.dtype,
            device=encoder_out.device,
        )[:, None, :]
        mask_chunk_predictor = None

        mask_chunk_predictor = self.encoder.overlap_chunk_cls.get_mask_chunk_predictor(
            None, device=encoder_out.device, batch_size=encoder_out.size(0)
        )
        mask_shfit_chunk = self.encoder.overlap_chunk_cls.get_mask_shfit_chunk(
            None, device=encoder_out.device, batch_size=encoder_out.size(0)
        )
        encoder_out = encoder_out * mask_shfit_chunk
        pre_acoustic_embeds, pre_token_length, pre_alphas, _ = self.predictor(
            encoder_out,
            ys_out_pad,
            encoder_out_mask,
            ignore_id=self.ignore_id,
            mask_chunk_predictor=mask_chunk_predictor,
            target_label_length=ys_in_lens,
        )
        predictor_alignments, predictor_alignments_len = (
            self.predictor.gen_frame_alignments(pre_alphas, encoder_out_lens)
        )

        encoder_chunk_size = self.encoder.overlap_chunk_cls.chunk_size_pad_shift_cur
        attention_chunk_center_bias = 0
        attention_chunk_size = encoder_chunk_size
        decoder_att_look_back_factor = (
            self.encoder.overlap_chunk_cls.decoder_att_look_back_factor_cur
        )
        mask_shift_att_chunk_decoder = (
            self.encoder.overlap_chunk_cls.get_mask_shift_att_chunk_decoder(
                None, device=encoder_out.device, batch_size=encoder_out.size(0)
            )
        )
        scama_mask = self.build_scama_mask_for_cross_attention_decoder_fn(
            predictor_alignments=predictor_alignments,
            encoder_sequence_length=encoder_out_lens,
            chunk_size=1,
            encoder_chunk_size=encoder_chunk_size,
            attention_chunk_center_bias=attention_chunk_center_bias,
            attention_chunk_size=attention_chunk_size,
            attention_chunk_type=self.decoder_attention_chunk_type,
            step=None,
            predictor_mask_chunk_hopping=mask_chunk_predictor,
            decoder_att_look_back_factor=decoder_att_look_back_factor,
            mask_shift_att_chunk_decoder=mask_shift_att_chunk_decoder,
            target_length=ys_in_lens,
            is_training=self.training,
        )

        return (
            pre_acoustic_embeds,
            pre_token_length,
            predictor_alignments,
            predictor_alignments_len,
            scama_mask,
        )

    def init_beam_search(
        self,
        **kwargs,
    ):

        from funasr_detach.models.scama.beam_search import BeamSearchScamaStreaming

        from funasr_detach.models.transformer.scorers.ctc import CTCPrefixScorer
        from funasr_detach.models.transformer.scorers.length_bonus import LengthBonus

        # 1. Build ASR model
        scorers = {}

        if self.ctc != None:
            ctc = CTCPrefixScorer(ctc=self.ctc, eos=self.eos)
            scorers.update(ctc=ctc)
        token_list = kwargs.get("token_list")
        scorers.update(
            decoder=self.decoder,
            length_bonus=LengthBonus(len(token_list)),
        )

        # 3. Build ngram model
        # ngram is not supported now
        ngram = None
        scorers["ngram"] = ngram

        weights = dict(
            decoder=1.0 - kwargs.get("decoding_ctc_weight", 0.0),
            ctc=kwargs.get("decoding_ctc_weight", 0.0),
            lm=kwargs.get("lm_weight", 0.0),
            ngram=kwargs.get("ngram_weight", 0.0),
            length_bonus=kwargs.get("penalty", 0.0),
        )

        beam_search = BeamSearchScamaStreaming(
            beam_size=kwargs.get("beam_size", 2),
            weights=weights,
            scorers=scorers,
            sos=self.sos,
            eos=self.eos,
            vocab_size=len(token_list),
            token_list=token_list,
            pre_beam_score_key=None if self.ctc_weight == 1.0 else "full",
        )
        # beam_search.to(device=kwargs.get("device", "cpu"), dtype=getattr(torch, kwargs.get("dtype", "float32"))).eval()
        # for scorer in scorers.values():
        #     if isinstance(scorer, torch.nn.Module):
        #         scorer.to(device=kwargs.get("device", "cpu"), dtype=getattr(torch, kwargs.get("dtype", "float32"))).eval()
        self.beam_search = beam_search

    def generate_chunk(
        self,
        speech,
        speech_lengths=None,
        key: list = None,
        tokenizer=None,
        frontend=None,
        **kwargs,
    ):
        cache = kwargs.get("cache", {})
        speech = speech.to(device=kwargs["device"])
        speech_lengths = speech_lengths.to(device=kwargs["device"])

        # Encoder
        encoder_out, encoder_out_lens = self.encode_chunk(
            speech, speech_lengths, cache=cache, is_final=kwargs.get("is_final", False)
        )
        if isinstance(encoder_out, tuple):
            encoder_out = encoder_out[0]
        if "running_hyps" not in cache:
            running_hyps = self.beam_search.init_hyp(encoder_out)
            cache["running_hyps"] = running_hyps

        # predictor
        predictor_outs = self.calc_predictor_chunk(
            encoder_out,
            encoder_out_lens,
            cache=cache,
            is_final=kwargs.get("is_final", False),
        )
        pre_acoustic_embeds, pre_token_length, alphas, pre_peak_index = (
            predictor_outs[0],
            predictor_outs[1],
            predictor_outs[2],
            predictor_outs[3],
        )
        pre_token_length = pre_token_length.round().long()

        if torch.max(pre_token_length) < 1:
            return []
        maxlen = minlen = pre_token_length
        if kwargs.get("is_final", False):
            maxlen += kwargs.get("token_num_relax", 5)
            minlen = max(0, minlen - kwargs.get("token_num_relax", 5))
        # c. Passed the encoder result and the beam search
        nbest_hyps = self.beam_search(
            x=encoder_out[0],
            scama_mask=None,
            pre_acoustic_embeds=pre_acoustic_embeds,
            maxlen=int(maxlen),
            minlen=int(minlen),
            cache=cache,
        )

        cache["running_hyps"] = nbest_hyps
        nbest_hyps = nbest_hyps[: self.nbest]

        results = []
        for hyp in nbest_hyps:
            # assert isinstance(hyp, (Hypothesis)), type(hyp)

            # remove sos/eos and get results
            last_pos = -1
            if isinstance(hyp.yseq, list):
                token_int = hyp.yseq[1:last_pos]
            else:
                token_int = hyp.yseq[1:last_pos].tolist()

                # remove blank symbol id, which is assumed to be 0
                token_int = list(
                    filter(
                        lambda x: x != self.eos
                        and x != self.sos
                        and x != self.blank_id,
                        token_int,
                    )
                )

                # Change integer-ids to tokens
                token = tokenizer.ids2tokens(token_int)
                # text = tokenizer.tokens2text(token)

                result_i = token

                results.extend(result_i)

        return results

    def init_cache(self, cache: dict = {}, **kwargs):
        device = kwargs.get("device", "cuda")

        chunk_size = kwargs.get("chunk_size", [0, 10, 5])
        encoder_chunk_look_back = kwargs.get("encoder_chunk_look_back", 0)
        decoder_chunk_look_back = kwargs.get("decoder_chunk_look_back", 0)
        batch_size = 1

        enc_output_size = kwargs["encoder_conf"]["output_size"]
        feats_dims = (
            kwargs["frontend_conf"]["n_mels"] * kwargs["frontend_conf"]["lfr_m"]
        )

        cache_encoder = {
            "start_idx": 0,
            "cif_hidden": torch.zeros((batch_size, 1, enc_output_size)).to(
                device=device
            ),
            "cif_alphas": torch.zeros((batch_size, 1)).to(device=device),
            "chunk_size": chunk_size,
            "encoder_chunk_look_back": encoder_chunk_look_back,
            "last_chunk": False,
            "opt": None,
            "feats": torch.zeros(
                (batch_size, chunk_size[0] + chunk_size[2], feats_dims)
            ).to(device=device),
            "tail_chunk": False,
        }
        cache["encoder"] = cache_encoder

        cache_decoder = {
            "decode_fsmn": None,
            "decoder_chunk_look_back": decoder_chunk_look_back,
            "opt": None,
            "chunk_size": chunk_size,
        }
        cache["decoder"] = cache_decoder
        cache["frontend"] = {}

        cache["prev_samples"] = torch.empty(0).to(device=device)

        return cache

    def inference(
        self,
        data_in,
        data_lengths=None,
        key: list = None,
        tokenizer=None,
        frontend=None,
        cache: dict = {},
        **kwargs,
    ):

        # init beamsearch
        is_use_ctc = (
            kwargs.get("decoding_ctc_weight", 0.0) > 0.00001 and self.ctc != None
        )
        is_use_lm = (
            kwargs.get("lm_weight", 0.0) > 0.00001
            and kwargs.get("lm_file", None) is not None
        )

        if self.beam_search is None:

            logging.info("enable beam_search")
            self.init_beam_search(**kwargs)
            self.nbest = kwargs.get("nbest", 1)

        if len(cache) == 0:
            self.init_cache(cache, **kwargs)

        meta_data = {}
        chunk_size = kwargs.get("chunk_size", [0, 10, 5])
        chunk_stride_samples = int(chunk_size[1] * 960)  # 600ms

        time1 = time.perf_counter()
        cfg = {"is_final": kwargs.get("is_final", False)}
        audio_sample_list = load_audio_text_image_video(
            data_in,
            fs=frontend.fs,
            audio_fs=kwargs.get("fs", 16000),
            data_type=kwargs.get("data_type", "sound"),
            tokenizer=tokenizer,
            cache=cfg,
        )
        _is_final = cfg["is_final"]  # if data_in is a file or url, set is_final=True

        time2 = time.perf_counter()
        meta_data["load_data"] = f"{time2 - time1:0.3f}"
        assert len(audio_sample_list) == 1, "batch_size must be set 1"

        audio_sample = torch.cat((cache["prev_samples"], audio_sample_list[0]))

        n = int(len(audio_sample) // chunk_stride_samples + int(_is_final))
        m = int(len(audio_sample) % chunk_stride_samples * (1 - int(_is_final)))
        tokens = []
        for i in range(n):
            kwargs["is_final"] = _is_final and i == n - 1
            audio_sample_i = audio_sample[
                i * chunk_stride_samples : (i + 1) * chunk_stride_samples
            ]

            # extract fbank feats
            speech, speech_lengths = extract_fbank(
                [audio_sample_i],
                data_type=kwargs.get("data_type", "sound"),
                frontend=frontend,
                cache=cache["frontend"],
                is_final=kwargs["is_final"],
            )
            time3 = time.perf_counter()
            meta_data["extract_feat"] = f"{time3 - time2:0.3f}"
            meta_data["batch_data_time"] = (
                speech_lengths.sum().item()
                * frontend.frame_shift
                * frontend.lfr_n
                / 1000
            )

            tokens_i = self.generate_chunk(
                speech,
                speech_lengths,
                key=key,
                tokenizer=tokenizer,
                cache=cache,
                frontend=frontend,
                **kwargs,
            )
            tokens.extend(tokens_i)

        text_postprocessed, _ = postprocess_utils.sentence_postprocess(tokens)

        result_i = {"key": key[0], "text": text_postprocessed}
        result = [result_i]

        cache["prev_samples"] = audio_sample[:-m]
        if _is_final:
            self.init_cache(cache, **kwargs)

        if kwargs.get("output_dir"):
            writer = DatadirWriter(kwargs.get("output_dir"))
            ibest_writer = writer[f"{1}best_recog"]
            ibest_writer["token"][key[0]] = " ".join(tokens)
            ibest_writer["text"][key[0]] = text_postprocessed

        return result, meta_data



# Copyright (c) 2024 Alibaba Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import random

import pyarrow.parquet as pq
import torch
import torchaudio
from torch.nn.utils.rnn import pad_sequence
import torch.nn.functional as F
import numpy as np
import re

torchaudio.set_audio_backend('soundfile')

AUDIO_FORMAT_SETS = {'flac', 'mp3', 'm4a', 'ogg', 'opus', 'wav', 'wma'}
CHORUS = {"intro": 0, "chorus": 1, "verse1": 2, "verse2": 3, "verse": 2,
		  "outro": 4}

metadata_pattern = re.compile(r'^\[(ti|ar|al|by|offset):.*\]$')
timestamp_pattern = re.compile(r'^\[\d{2}:\d{2}\.\d{2}\](.*)$')


def parquet_opener(data, mode='train', audio_data={}):
	""" Give url or local file, return file descriptor
        Inplace operation.

        Args:
            data(Iterable[str]): url or local file list

        Returns:
            Iterable[{src, stream}]
    """
	for sample in data:
		assert 'src' in sample

		url = sample['src']
		try:
			df = pq.read_table(url).to_pandas()
			for i in df.index:
				sample.update(dict(df.loc[i]))
				yield {**sample}
		except Exception as ex:
			logging.warning('Failed to open {}, ex info {}'.format(url, ex))


def clean_lyrics(data, mode="train"):
	for sample in data:
		lyrics = sample["text"]
		cleaned = []
		for line in lyrics.splitlines():
			if metadata_pattern.match(line):
				continue
			timestamp_match = timestamp_pattern.match(line)
			if timestamp_match:
				lyric = timestamp_match.group(1).strip()
				if lyric:
					cleaned.append(lyric)
			else:
				if line.strip():
					cleaned.append(line.strip())
		sample["text"] = '\n'.join(cleaned)
		yield sample


def cut_by_length(data, max_length=8000, num_times=4, mode="train"):
	for sample in data:
		if "semantic_token" in sample:
			sample["semantic_token"] = [
				sample["semantic_token"][0][:max_length]]
		if "acoustic_token" not in sample:
			sample["acoustic_token"] = sample["speech_token"]
		sample["acoustic_token"] = sample["acoustic_token"][
								   :max_length * num_times]

		yield sample


def filter(data,
           max_length=22500,  # 22500 #5min #10240
           max_acoustic_length=45000,
           min_length=10,
           min_acoustic_length=150,
           token_max_length=200,
           token_min_length=1,
           min_output_input_ratio=0.0005,
           max_output_input_ratio=1,
           mode='train'):
	""" Filter sample according to feature and label length
        Inplace operation.

        Args::
            data: Iterable[{key, wav, label, sample_rate}]
            max_length: drop utterance which is greater than max_length(10ms)
            min_length: drop utterance which is less than min_length(10ms)
            token_max_length: drop utterance which is greater than
                token_max_length, especially when use char unit for
                english modeling
            token_min_length: drop utterance which is
                less than token_max_length
            min_output_input_ratio: minimal ration of
                token_length / feats_length(10ms)
            max_output_input_ratio: maximum ration of
                token_length / feats_length(10ms)

        Returns:
            Iterable[{key, wav, label, sample_rate}]
    """
	if mode == "train":
		for sample in data:
			if "semantic_token" in sample:
				new_sample_frames = sample['semantic_token'][0].shape[0]
			else:
				new_sample_frames = sample['speech_token']

			if "text_token" in sample:
				new_sample_frames += len(sample['text_token'])

			if new_sample_frames > max_length or new_sample_frames < min_length:
				print(f"skipped 1 item length={new_sample_frames}")
				continue

			sample["chorus"] = sample["chorus"].split(",")
			if not isinstance(sample["time_start"], np.ndarray):
				sample["time_start"] = [sample["time_start"]]
				sample["time_end"] = [sample["time_end"]]
			for i, t in enumerate(sample["chorus"]):
				if sample["chorus"][i] == "verse":
					sample["chorus"][i] = "verse1"

			yield sample

	if mode == "train_flow":
		for sample in data:
			if "semantic_token" in sample:
				new_sample_frames = sample['semantic_token'][0].shape[0]
			if "acoustic_token" in sample:
				target_sample_frames = sample['acoustic_token'][0].shape[0]

			if new_sample_frames > max_length or new_sample_frames < min_acoustic_length or new_sample_frames < min_length or target_sample_frames > max_acoustic_length:
				print(
					f"skipped 1 item length={new_sample_frames}, target_length={target_sample_frames}")
				continue

			yield sample

	elif mode == "inference":
		for sample in data:
			yield sample


def resample(data, resample_rate=22050, min_sample_rate=16000, mode='train'):
	""" Resample data.
        Inplace operation.

        Args:
            data: Iterable[{key, wav, label, sample_rate}]
            resample_rate: target resample rate

        Returns:
            Iterable[{key, wav, label, sample_rate}]
    """
	for sample in data:
		assert 'sample_rate' in sample
		assert 'speech' in sample
		sample_rate = sample['sample_rate']
		waveform = sample['speech']
		if sample_rate != resample_rate:
			if sample_rate < min_sample_rate:
				continue
			sample['sample_rate'] = resample_rate
			sample['speech'] = torchaudio.transforms.Resample(
					orig_freq=sample_rate, new_freq=resample_rate)(waveform)
		max_val = sample['speech'].abs().max()
		if max_val > 1:
			sample['speech'] /= max_val
		yield sample


def truncate(data, truncate_length=24576, mode='train'):
	""" Truncate data.

        Args:
            data: Iterable[{key, wav, label, sample_rate}]
            truncate_length: truncate length

        Returns:
            Iterable[{key, wav, label, sample_rate}]
    """
	for sample in data:
		waveform = sample['audio']
		if waveform.shape[1] > truncate_length:
			start = random.randint(0, waveform.shape[1] - truncate_length)
			waveform = waveform[:, start: start + truncate_length]
		else:
			waveform = torch.concat([waveform, torch.zeros(1, truncate_length -
														   waveform.shape[1])],
									dim=1)
		sample['audio'] = waveform
		yield sample


def upsample(data, resample_rate=48000, min_sample_rate=16000, mode='train',
			 n_codebook=4):
	""" Resample data.
        Inplace operation.

        Args:
            data: Iterable[{key, wav, label, sample_rate}]
            resample_rate: target resample rate

        Returns:
            Iterable[{key, wav, label, sample_rate}]
    """
	for sample in data:
		assert 'semantic_token' in sample
		# TODO: unify data processing key names
		if 'acoustic_token' not in sample:
			continue

		if 'sample_rate' in sample.keys():
			sample_rate = sample['sample_rate']
		else:
			sample_rate = 24000
		token = np.array(sample['semantic_token'][0][:-1])

		# Calculate the repetition factor for resampling
		repetition_factor = int(n_codebook * resample_rate / sample_rate)
		if sample_rate != resample_rate:
			if sample_rate < min_sample_rate:
				continue
			sample['sample_rate'] = resample_rate
			sample['semantic_token'] = np.array(
					[np.repeat(token, repetition_factor)])

		yield sample

def compute_fbank(data,
				  feat_extractor,
				  mode='train'):
	""" Extract fbank

        Args:
            data: Iterable[{key, wav, label, sample_rate}]

        Returns:
            Iterable[{key, feat, label}]
    """
	for sample in data:
		assert 'sample_rate' in sample
		assert 'speech' in sample
		assert 'utt' in sample
		assert 'text_token' in sample
		waveform = sample['speech']
		mat = feat_extractor(waveform).squeeze(dim=0).transpose(0, 1)
		sample['speech_feat'] = mat
		del sample['speech']
		yield sample


def parse_embedding(data, normalize, mode='train'):
	""" Parse utt_embedding/spk_embedding

        Args:
            data: Iterable[{key, wav, label, sample_rate}]

        Returns:
            Iterable[{key, feat, label}]
    """

	for sample in data:
		sample['utt_embedding'] = torch.tensor(sample['utt_embedding'],
											   dtype=torch.float32)
		sample['spk_embedding'] = torch.tensor(sample['spk_embedding'],
											   dtype=torch.float32)
		if normalize:
			sample['utt_embedding'] = F.normalize(sample['utt_embedding'],
												  dim=0)
			sample['spk_embedding'] = F.normalize(sample['spk_embedding'],
												  dim=0)
		yield sample

def tokenize(data, get_tokenizer, allowed_special, mode='train'):
	""" Decode text to chars or BPE
        Inplace operation

        Args:
            data: Iterable[{key, wav, txt, sample_rate}]

        Returns:
            Iterable[{key, wav, txt, tokens, label, sample_rate}]
    """
	tokenizer = get_tokenizer()

	for sample in data:
		assert 'text' in sample
		sample['text_token'] = tokenizer.encode(sample['text'],
												allowed_special=allowed_special)
		yield sample


def shuffle(data, shuffle_size=10000, mode='train'):
	""" Local shuffle the data

        Args:
            data: Iterable[{key, feat, label}]
            shuffle_size: buffer size for shuffle

        Returns:
            Iterable[{key, feat, label}]
    """
	buf = []
	for sample in data:
		buf.append(sample)
		if len(buf) >= shuffle_size:
			random.shuffle(buf)
			for x in buf:
				yield x
			buf = []
	# The sample left over
	random.shuffle(buf)
	for x in buf:
		yield x


def sort(data, sort_size=500, mode='train'):
	""" Sort the data by feature length.
        Sort is used after shuffle and before batch, so we can group
        utts with similar lengths into a batch, and `sort_size` should
        be less than `shuffle_size`

        Args:
            data: Iterable[{key, feat, label}]
            sort_size: buffer size for sort

        Returns:
            Iterable[{key, feat, label}]
    """

	buf = []
	for sample in data:
		if sample["chorus"] == "verse":
			sample["chorus"] = "verse1"

		if sample["acoustic_token"].shape[0] == 1:
			sample["acoustic_token"] = np.concatenate(
					sample["acoustic_token"][0])
		else:
			sample["acoustic_token"] = np.concatenate(sample["acoustic_token"])

		sample["acoustic_token"] = torch.from_numpy(sample["acoustic_token"])
		buf.append(sample)
		if len(buf) >= sort_size:
			buf.sort(key=lambda x: x['acoustic_token'].size(0))
			for x in buf:
				yield x
			buf = []
	# The sample left over
	buf.sort(key=lambda x: x['acoustic_token'].size(0))
	for x in buf:
		yield x


def static_batch(data, batch_size=32):
	""" Static batch the data by `batch_size`

        Args:
            data: Iterable[{key, feat, label}]
            batch_size: batch size

        Returns:
            Iterable[List[{key, feat, label}]]
    """
	buf = []
	data_empty = True
	for sample in data:
		data_empty = False
		buf.append(sample)
		if len(buf) >= batch_size:
			yield buf
			buf = []
	if data_empty:
		raise ValueError("data is empty")
	if len(buf) > 0:
		yield buf


def dynamic_batch(data, max_frames_in_batch=12000, mode='train'):
	""" Dynamic batch the data until the total frames in batch
        reach `max_frames_in_batch`

        Args:
            data: Iterable[{key, feat, label}]
            max_frames_in_batch: max_frames in one batch

        Returns:
            Iterable[List[{key, feat, label}]]
    """
	buf = []
	longest_frames = 0
	for sample in data:
		assert 'acoustic_token' in sample
		assert isinstance(sample['acoustic_token'], torch.Tensor)

		if 'semantic_token' in sample:
			new_sample_frames = sample['semantic_token'][0].shape[0]
		else:
			new_sample_frames = sample['semantic_token']

		if "text_token" in sample:
			new_sample_frames += len(sample['text_token'])

		longest_frames = max(longest_frames, new_sample_frames)
		frames_after_padding = longest_frames * (len(buf) + 1)

		if frames_after_padding > max_frames_in_batch:
			if len(buf) > 0:
				yield buf
			buf = [sample]
			longest_frames = new_sample_frames
		else:
			buf.append(sample)
	if len(buf) > 0:
		yield buf


def batch(data, batch_type='static', batch_size=16, max_frames_in_batch=12000,
		  mode='train'):
	""" Wrapper for static/dynamic batch
    """
	if mode == 'inference':
		return static_batch(data, 1)
	elif mode == 'processing':
		return static_batch(data, batch_size)
	else:
		if batch_type == 'static':
			return static_batch(data, batch_size)
		elif batch_type == 'dynamic':
			return dynamic_batch(data, max_frames_in_batch)
		else:
			logging.fatal('Unsupported batch type {}'.format(batch_type))


def padding(data, mode='train'):
	""" Padding the data into training data

        Args:
            data: Iterable[List[{key, feat, label}]]

        Returns:
            Iterable[Tuple(keys, feats, labels, feats lengths, label lengths)]
    """
	if mode == "train":
		for sample in data:
			assert isinstance(sample, list)
			if len(sample) != 0:
				acoustic_feat_len = torch.tensor(
						[x['acoustic_token'].size(0) for x in sample],
						dtype=torch.int32)
				order = torch.argsort(acoustic_feat_len, descending=True)
				utts = [sample[i]['utt'] for i in order]
				acoustic_token = [
					sample[i]['acoustic_token'].clone().to(torch.int32) for i in
					order]
				acoustic_token_len = torch.tensor(
						[i.size(0) for i in acoustic_token], dtype=torch.int32)

				acoustic_token = pad_sequence(acoustic_token,
											  batch_first=True,
											  padding_value=0)

				text = [sample[i]['text'] for i in order]
				text_token = [torch.tensor(sample[i]['text_token']).long() for i
							  in order]
				text_token_len = torch.tensor([i.size(0) for i in text_token],
											  dtype=torch.int32)
				text_token = pad_sequence(text_token, batch_first=True,
										  padding_value=0)
				time_start = torch.tensor(
						[sample[i]['time_start'] for i in order])
				time_end = torch.tensor([sample[i]['time_end'] for i in order])

				if isinstance(sample[0]['chorus'], str):
					chorus = torch.tensor(
							[CHORUS[sample[i]['chorus']] for i in order])
				else:
					chorus = [
						torch.tensor([CHORUS[t] for t in sample[i]['chorus']])
						for i in order]
					chorus = pad_sequence(chorus, batch_first=True,
										  padding_value=-1)

				batch = {
					"utts"              : utts,
					"acoustic_token"    : acoustic_token,
					"acoustic_token_len": acoustic_token_len,
					"time_start"        : time_start,
					"time_end"          : time_end,
					"chorus"            : chorus,
					"text"              : text,
					"text_token"        : text_token,
					"text_token_len"    : text_token_len,
				}

				if "semantic_token" in sample[0]:
					semantic_token = [
						torch.tensor(sample[i]['semantic_token'][0],
									 dtype=torch.int32) for i in order]
					semantic_token_len = torch.tensor(
							[i.size(0) for i in semantic_token],
							dtype=torch.int32)
					semantic_token = pad_sequence(semantic_token,
												  batch_first=True,
												  padding_value=0)
					batch.update({"semantic_token"    : semantic_token,
								  "semantic_token_len": semantic_token_len})

				yield batch
			else:
				logging.info("WARNING: sample is empty []!")

	elif mode == "inference":
		for sample in data:
			assert isinstance(sample, list)
			utts = [sample[i]['utt'] for i in range(len(sample))]
			text = [sample[i]['text'] for i in range(len(sample))]
			text_token = [torch.tensor(sample[i]['text_token']).long() for i in
						  range(len(sample))]
			text_token_len = torch.tensor([i.size(0) for i in text_token],
										  dtype=torch.int32)
			text_token = pad_sequence(text_token, batch_first=True,
									  padding_value=0)
			time_start = torch.tensor(
					[sample[i]['time_start'] for i in range(len(sample))])
			time_end = torch.tensor(
					[sample[i]['time_end'] for i in range(len(sample))])

			if isinstance(sample[0]['chorus'], str):
				chorus = torch.tensor([CHORUS[sample[i]['chorus']] for i in
									   range(len(sample))])
			else:
				chorus = [torch.tensor([CHORUS[t] for t in sample[i]['chorus']])
						  for i in range(len(sample))]
				chorus = pad_sequence(chorus, batch_first=True,
									  padding_value=-1)

			if "acoustic_token" in sample[0]:
				acoustic_token = [
					sample[i]['acoustic_token'].clone().to(torch.int32) for i in
					range(len(sample))]
				acoustic_token_len = torch.tensor(
						[i.size(0) for i in acoustic_token], dtype=torch.int32)
				acoustic_token = pad_sequence(acoustic_token,
											  batch_first=True,
											  padding_value=0)
			else:
				acoustic_token = None
				acoustic_token_len = None

			batch = {
				"utts"              : utts,
				"acoustic_token"    : acoustic_token,
				"acoustic_token_len": acoustic_token_len,
				"time_start"        : time_start,
				"time_end"          : time_end,
				"chorus"            : chorus,
				"text"              : text,
				"text_token"        : text_token,
				"text_token_len"    : text_token_len,
			}

			if "semantic_token" in sample[0]:
				semantic_token = [torch.tensor(sample[i]['semantic_token'][0],
											   dtype=torch.int32) for i in
								  range(len(sample))]
				semantic_token_len = torch.tensor(
						[i.size(0) for i in semantic_token], dtype=torch.int32)
				semantic_token = pad_sequence(semantic_token,
											  batch_first=True,
											  padding_value=0)
				batch.update({"semantic_token"    : semantic_token,
							  "semantic_token_len": semantic_token_len})

			yield batch



from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.runner import PipelineRunner
from apipeline.pipeline.task import PipelineTask
from apipeline.processors.logger import FrameLogger

from achatbot.processors.vision.detect_processor import DetectProcessor
from achatbot.processors.speech.tts.tts_processor import TTSProcessor
from achatbot.cmd.bots.base_agora import AgoraChannelBot
from achatbot.common.types import AgoraParams
from achatbot.transports.agora import AgoraTransport
from .. import register_ai_room_bots


@register_ai_room_bots.register
class AgoraDetectVisionBot(AgoraChannelBot):
    def __init__(self, **args) -> None:
        super().__init__(**args)
        self.init_bot_config()

    async def arun(self):
        agora_params = AgoraParams(
            camera_in_enabled=True,
            audio_out_enabled=True,
        )
        detect_processor: DetectProcessor = self.get_vision_detect_processor()
        tts_processor: TTSProcessor = self.get_tts_processor()
        stream_info = tts_processor.get_stream_info()
        agora_params.audio_out_sample_rate = stream_info["sample_rate"]
        agora_params.audio_out_channels = stream_info["channels"]
        transport = AgoraTransport(self.args.token, params=agora_params)

        @transport.event_handler("on_first_participant_joined")
        async def on_first_participant_joined(
            transport: AgoraTransport,
            user_id: str,
        ):
            # subscribed the first participant
            transport.capture_participant_video(user_id)

            participant_name = user_id
            await tts_processor.say(
                f"你好,{participant_name}。"
                f"这是一个图像检测hello demo。"
                f"当检测到条件对象时，说欢迎词。"
                f"当未检测到条件对象时，说离开词。"
            )

        pipeline = Pipeline(
            [
                transport.input_processor(),
                detect_processor,
                tts_processor,
                # FrameLogger(include_frame_types=[UserImageRawFrame]),
                transport.output_processor(),
            ]
        )
        self.task = PipelineTask(pipeline)
        await PipelineRunner().run(self.task)
